#
# -------------------------------------------------------------
#    Copyright 2004-2011 Synopsys, Inc.
#    Copyright 2010-2011 Mentor Graphics Corporation
#    Copyright 2010-2011 Cadence Design Systems, Inc.
#    Copyright 2019 Tuomas Poikela
#    All Rights Reserved Worldwide
#
#    Licensed under the Apache License, Version 2.0 (the
#    "License"); you may not use this file except in
#    compliance with the License.  You may obtain a copy of
#    the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in
#    writing, software distributed under the License is
#    distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#    CONDITIONS OF ANY KIND, either express or implied.  See
#    the License for the specific language governing
#    permissions and limitations under the License.
#
#    Original SV code has been adapter to Python.
# -------------------------------------------------------------

#import cocotb

from typing import Dict, Optional, List

from ..base.sv import sv
from ..base.uvm_object import UVMObject
from ..base.uvm_globals import uvm_check_output_args
from ..base.uvm_object_globals import (UVM_HIGH)
from ..base.uvm_pool import UVMPool, UVMObjectStringPool
from ..base.uvm_queue import UVMQueue
from ..base.uvm_resource_db import UVMResourceDb
from .uvm_reg_model import (UVM_IS_OK, UVM_HAS_X, UVM_NO_COVERAGE,
    UVM_DEFAULT_PATH, UVM_HIER, uvm_reg_cvr_rsrc_db, UVM_FRONTDOOR,
    UVM_NO_CHECK, UVM_CVR_ALL, UVM_NOT_OK)
from .uvm_reg_map import UVMRegMap
from .uvm_mem import UVMMem
from .uvm_reg_field import UVMRegField
from .uvm_reg import UVMReg
from ..macros import (uvm_info, uvm_error, uvm_warning, uvm_fatal, UVM_REG_DATA_WIDTH)

ERR_MSG1 = ("There are %0d root register models named %s. The names of the root"
    + " register models have to be unique")

ERR_MSG3 = ("Register model requires that UVM_REG_DATA_WIDTH be defined as %0d"
    + " or greater. Currently defined as %0d")


class UVMRegBlock(UVMObject):
    """
    Block abstraction base class

    A block represents a design hierarchy. It can contain registers,
    register files, memories and sub-blocks.

    A block has one or more address maps, each corresponding to a physical
    interface on the block.
    """

    m_roots: Dict['UVMRegBlock', int] = {}
    id = 0


    #   //----------------------
    #   // Group: Initialization
    #   //----------------------

    #   // Function: new
    #   //
    #   // Create a new instance and type-specific configuration
    #   //
    #   // Creates an instance of a block abstraction class with the specified
    #   // name.
    #   //
    #   // ~has_coverage~ specifies which functional coverage models are present in
    #   // the extension of the block abstraction class.
    #   // Multiple functional coverage models may be specified by adding their
    #   // symbolic names, as defined by the <uvm_coverage_model_e> type.
    #   //
    #   function new(string name="", int has_coverage=UVM_NO_COVERAGE)
    def __init__(self, name="", has_coverage=UVM_NO_COVERAGE):
        super().__init__(name)
        self.hdl_paths_pool = UVMObjectStringPool("hdl_paths", UVMQueue)
        self.has_cover = has_coverage
        self.cover_on = 0
        # Root block until registered with a parent
        UVMRegBlock.m_roots[self] = 0
        self.locked = False
        self.maps = UVMPool()  # bit[UVMRegMap]
        self.regs = UVMPool()  # int unsigned[uvm_reg]
        self.blks = UVMPool()
        self.mems = UVMPool()
        self.vregs = UVMPool()
        self.root_hdl_paths = UVMPool()
        self.backdoor = None
        self.default_hdl_path = "RTL"
        self.parent = None
        #   // Variable: default_path
        #   // Default access path for the registers and memories in this block.
        self.default_path = UVM_DEFAULT_PATH
        self.fname = ""
        self.lineno = 0


    #   // Function: configure
    #   //
    #   // Instance-specific configuration
    #   //
    #   // Specify the parent block of this block.
    #   // A block without parent is a root block.
    #   //
    #   // If the block file corresponds to a hierarchical RTL structure,
    #   // its contribution to the HDL path is specified as the ~hdl_path~.
    #   // Otherwise, the block does not correspond to a hierarchical RTL
    #   // structure (e.g. it is physically flattened) and does not contribute
    #   // to the hierarchical HDL path of any contained registers or memories.
    #   //
    def configure(self, parent=None, hdl_path=""):
        self.parent = parent
        if parent is not None:
            self.parent.add_block(self)
        self.add_hdl_path(hdl_path)

        UVMResourceDb.set("uvm_reg::*", self.get_full_name(), self)


    #   // Function: create_map
    #   //
    #   // Create an address map in this block
    #   //
    #   // Create an address map with the specified ~name~, then
    #   // configures it with the following properties.
    #   //
    #   // base_addr - the base address for the map. All registers, memories,
    #   //             and sub-blocks within the map will be at offsets to this
    #   //             address
    #   //
    #   // n_bytes   - the byte-width of the bus on which this map is used
    #   //
    #   // endian    - the endian format. See <uvm_endianness_e> for possible
    #   //             values
    #   //
    #   // byte_addressing - specifies whether consecutive addresses refer are 1 byte
    #   //             apart (TRUE) or ~n_bytes~ apart (FALSE). Default is TRUE.
    #   //
    #   //| APB = create_map("APB", 0, 1, UVM_LITTLE_ENDIAN, 1)
    #   //
    #   extern virtual function UVMRegMap create_map(string name,
    #                                                  uvm_reg_addr_t base_addr,
    #                                                  int unsigned n_bytes,
    #                                                  uvm_endianness_e endian,
    #                                                  bit byte_addressing = 1)
    def create_map(self, name: str, base_addr: int, n_bytes: int, endian: int,
            byte_addressing=True) -> Optional[UVMRegMap]:
        #   UVMRegMap  map
        if self.locked is True:
            uvm_error("RegModel", "Cannot add map to locked model")
            return None

        _map = UVMRegMap.type_id.create(name, None, self.get_full_name())
        _map.configure(self,base_addr,n_bytes,endian,byte_addressing)

        self.maps[_map] = True
        if self.maps.num() == 1:
            self.default_map = _map
        return _map


    #   // Function: check_data_width
    #   //
    #   // Check that the specified data width (in bits) is less than
    #   // or equal to the value of `UVM_REG_DATA_WIDTH
    #   //
    #   // This method is designed to be called by a static initializer
    #   //
    #   //| class my_blk extends uvm_reg_block
    #   //|   local static bit m_data_width = check_data_width(356)
    #   //|   ...
    #   //| endclass
    #   //
    def check_data_width(self, width: int) -> int:
        # if width <= sv.bits(uvm_reg_data_t):
        if width <= UVM_REG_DATA_WIDTH:
            return 1

        uvm_fatal("RegModel", sv.sformatf(
            "Register model requires that UVM_REG_DATA_WIDTH be defined as %0d or greater. Currently defined as %0d",
            width, UVM_REG_DATA_WIDTH))
        return 0


    #   // Function: set_default_map
    #   //
    #   // Defines the default address map
    #   //
    #   // Set the specified address map as the <default_map> for this
    #   // block. The address map must be a map of this address block.
    #   //
    #   extern function void set_default_map (UVMRegMap map)
    def set_default_map(self, _map: UVMRegMap):
        if _map not in self.maps:
            uvm_warning("RegModel", "Map '" + _map.get_full_name() + "' does not exist in block")
        self.default_map = _map


    #   // Variable: default_map
    #   //
    #   // Default address map
    #   //
    #   // Default address map for this block, to be used when no
    #   // address map is specified for a register operation and that
    #   // register is accessible from more than one address map.
    #   //
    #   // It is also the implicit address map for a block with a single,
    #   // unnamed address map because it has only one physical interface.
    #   //
    #   UVMRegMap default_map

    #
    #   extern function UVMRegMap get_default_map ()
    #
    def get_default_map(self) -> UVMRegMap:
        return self.default_map

    #   extern virtual function void set_parent(uvm_reg_block parent)
    def set_parent(self, parent: 'UVMRegBlock'):
        if self != parent:
            self.parent = parent


    # function void add_block (uvm_reg_block blk)
    def add_block(self, blk: 'UVMRegBlock'):
        if self.is_locked():
            uvm_error("RegModel", "Cannot add subblock to locked block model")
            return

        if blk in self.blks:
            uvm_error("RegModel", "Subblock '" + blk.get_name()
                    + "' has already been registered with block '" + self.get_name() + "'")
            return

        self.blks[blk] = UVMRegBlock.id
        UVMRegBlock.id += 1
        if blk in UVMRegBlock.m_roots:
            del UVMRegBlock.m_roots[blk]


    #   /*local*/ extern function void add_map   (UVMRegMap map)
    def add_map(self, _map: UVMRegMap):
        if (self.locked):
            uvm_error("RegModel", "Cannot add map to locked model")
            return

        if self.maps.exists(_map):
            uvm_error("RegModel", "Map '" + _map.get_name()
                      + "' already exists in '" + self.get_full_name() + "'")
            return
        self.maps[_map] = 1
        if self.maps.num() == 1:
            self.default_map = _map


    # add_reg
    def add_reg(self, rg: UVMReg):
        if self.is_locked():
            uvm_error("RegModel", "Cannot add register to locked block model")
            return

        if rg in self.regs:
            uvm_error("RegModel", ("Register '" + rg.get_name()
              + "' has already been registered with block '"
              + self.get_name() + "'"))
            return

        self.regs[rg] = UVMRegBlock.id
        UVMRegBlock.id += 1

    #   /*local*/ extern function void add_vreg  (uvm_vreg vreg)
    def add_vreg(self, vreg):
        if self.is_locked():
            uvm_error("RegModel", "Cannot add virtual register to locked block model")
            return
        
        if self.vregs.exists(vreg):
            uvm_error("RegModel", ("Virtual register '" + vreg.get_name() +
                "' has already been registered with block '" + self.get_name() + "'"))
            return
        self.vregs[vreg] = UVMRegBlock.id
        UVMRegBlock.id += 1

    #   /*local*/ extern function void add_mem   (uvm_mem  mem)
    def add_mem(self, mem: UVMMem):
        if self.is_locked():
            uvm_error("RegModel", "Cannot add memory to locked block model")
            return

        if mem in self.mems:
            uvm_error("RegModel", "Memory '" + mem.get_name()
                + "' has already been registered with block '" + self.get_name()
                + "'")
            return

        self.mems[mem] = UVMRegBlock.id
        UVMRegBlock.id += 1


    #   // Function: lock_model
    #   //
    #   // Lock a model and build the address map.
    #   //
    #   // Recursively lock an entire register model
    #   // and build the address maps to enable the
    #   // <UVMRegMap::get_reg_by_offset()> and
    #   // <UVMRegMap::get_mem_by_offset()> methods.
    #   //
    #   // Once locked, no further structural changes,
    #   // such as adding registers or memories,
    #   // can be made.
    #   //
    #   // It is not possible to unlock a model.
    #   //
    def lock_model(self):
        if self.is_locked():
            return
        self.locked = 1

        for rr in self.regs.key_list():
            rr.Xlock_modelX()

        for mem in self.mems.key_list():
            mem.Xlock_modelX()

        for blk in self.blks.key_list():
            blk.lock_model()

        # TODO finish this
        if (self.parent is None):
            max_size = UVMReg.get_max_size()
            if (UVMRegField.get_max_size() > max_size):
                max_size = UVMRegField.get_max_size()

            if (UVMMem.get_max_size() > max_size):
                max_size = UVMMem.get_max_size()

            if (max_size > UVM_REG_DATA_WIDTH):
                uvm_fatal("RegModel", sv.sformatf(ERR_MSG3, max_size, UVM_REG_DATA_WIDTH))

            self.Xinit_address_mapsX()

            # Check that root register models have unique names

            # Has this name has been checked before?
            if self in UVMRegBlock.m_roots and UVMRegBlock.m_roots[self] != 1:
                n = 0

                for _blk in UVMRegBlock.m_roots:
                    blk = _blk  # uvm_reg_block
                    if (blk.get_name() == self.get_name()):
                        UVMRegBlock.m_roots[blk] = 1
                        n += 1

                if (n > 1):
                    uvm_error("UVM/REG/DUPLROOT", sv.sformatf(ERR_MSG1, n, self.get_name()))


    #   // Function: is_locked
    #   //
    #   // Return TRUE if the model is locked.
    #   //
    def is_locked(self):
        return self.locked

    #   //---------------------
    #   // Group: Introspection
    #   //---------------------
    #
    #
    #   // Function: get_name
    #   //
    #   // Get the simple name
    #   //
    #   // Return the simple object name of this block.
    #   //



    #   // Function: get_full_name
    #   //
    #   // Get the hierarchical name
    #   //
    #   // Return the hierarchal name of this block.
    #   // The base of the hierarchical name is the root block.
    #   //
    #   extern virtual function string get_full_name()
    def get_full_name(self) -> str:
        if self.parent is None:
            return self.get_name()

        return self.parent.get_full_name() + "." + self.get_name()


    #   // Function: get_parent
    #   //
    #   // Get the parent block
    #   //
    #   // If this a top-level block, returns ~None~.
    #   //
    #   extern virtual function uvm_reg_block get_parent()
    def get_parent(self) -> Optional['UVMRegBlock']:
        return self.parent


    #   // Function: get_root_blocks
    #   //
    #   // Get the all root blocks
    #   //
    #   // Returns an array of all root blocks in the simulation.
    #   //
    #   extern static  function void get_root_blocks(ref uvm_reg_block blks[$])
    def get_root_blocks(self, blks):
        for blk in UVMRegBlock.m_roots:
            blks.append(blk)


    #   // Function: find_blocks
    #   //
    #   // Find the blocks whose hierarchical names match the
    #   // specified ~name~ glob.
    #   // If a ~root~ block is specified, the name of the blocks are
    #   // relative to that block, otherwise they are absolute.
    #   //
    #   // Returns the number of blocks found.
    #   //
    #   extern static function int find_blocks(input string        name,
    #                                          ref   uvm_reg_block blks[$],
    #                                          input uvm_reg_block root = None,
    #                                          input uvm_object    accessor = None)


    #   // Function: find_block
    #   //
    #   // Find the first block whose hierarchical names match the
    #   // specified ~name~ glob.
    #   // If a ~root~ block is specified, the name of the blocks are
    #   // relative to that block, otherwise they are absolute.
    #   //
    #   // Returns the first block found or ~None~ otherwise.
    #   // A warning is issued if more than one block is found.
    #   //
    #   extern static function uvm_reg_block find_block(input string        name,
    #                                                   input uvm_reg_block root = None,
    #                                                   input uvm_object    accessor = None)



    #   // Function: get_blocks
    #   //
    #   // Get the sub-blocks
    #   //
    #   // Get the blocks instantiated in this blocks.
    #   // If ~hier~ is TRUE, recursively includes any sub-blocks.
    #   //
    #   extern virtual function void get_blocks (ref uvm_reg_block  blks[$],
    #                                            input uvm_hier_e hier=UVM_HIER)
    def get_blocks(self, blks, hier=UVM_HIER):
        for blk_ in self.blks.key_list():
            blk = blk_
            blks.append(blk)
            if hier == UVM_HIER:
                blk.get_blocks(blks)


    #   // Function: get_maps
    #   //
    #   // Get the address maps
    #   //
    #   // Get the address maps instantiated in this block.
    #   //
    #   extern virtual function void get_maps (ref UVMRegMap maps[$])
    def get_maps(self, maps):
        for key in self.maps.key_list():
            maps.append(key)


    #   // Function: get_registers
    #   //
    #   // Get the registers
    #   //
    #   // Get the registers instantiated in this block.
    #   // If ~hier~ is TRUE, recursively includes the registers
    #   // in the sub-blocks.
    #   //
    #   // Note that registers may be located in different and/or multiple
    #   // address maps. To get the registers in a specific address map,
    #   // use the <UVMRegMap::get_registers()> method.
    #   //
    #   extern virtual function void get_registers (ref uvm_reg regs[$],
    #                                               input uvm_hier_e hier=UVM_HIER)
    #
    #
    def get_registers(self, regs, hier=UVM_HIER):
        for rg in self.regs.key_list():
            regs.append(rg)

        if hier == UVM_HIER:
            for blk_ in self.blks.key_list():
                blk = blk_
                blk.get_registers(regs)


    #   // Function: get_fields
    #   //
    #   // Get the fields
    #   //
    #   // Get the fields in the registers instantiated in this block.
    #   // If ~hier~ is TRUE, recursively includes the fields of the registers
    #   // in the sub-blocks.
    #   //
    #   extern virtual function void get_fields (ref uvm_reg_field  fields[$],
    #                                            input uvm_hier_e hier=UVM_HIER)


    #   // Function: get_memories
    #   //
    #   // Get the memories
    #   //
    #   // Get the memories instantiated in this block.
    #   // If ~hier~ is TRUE, recursively includes the memories
    #   // in the sub-blocks.
    #   //
    #   // Note that memories may be located in different and/or multiple
    #   // address maps. To get the memories in a specific address map,
    #   // use the <UVMRegMap::get_memories()> method.
    #   //
    #   extern virtual function void get_memories (ref uvm_mem mems[$],
    #                                              input uvm_hier_e hier=UVM_HIER)
    def get_memories(self, mems, hier=UVM_HIER):
        #   foreach (self.mems[mem_]):
        for mem_ in self.mems.key_list():
            mem = mem_
            mems.append(mem)

        if hier == UVM_HIER:
            for blk_ in self.blks.key_list():
                blk = blk_
                blk.get_memories(mems)



    #   // Function: get_virtual_registers
    #   //
    #   // Get the virtual registers
    #   //
    #   // Get the virtual registers instantiated in this block.
    #   // If ~hier~ is TRUE, recursively includes the virtual registers
    #   // in the sub-blocks.
    #   //
    #   extern virtual function void get_virtual_registers(ref uvm_vreg regs[$],
    #                                                input uvm_hier_e hier=UVM_HIER)

    #   // Function: get_virtual_fields
    #   //
    #   // Get the virtual fields
    #   //
    #   // Get the virtual fields from the virtual registers instantiated
    #   // in this block.
    #   // If ~hier~ is TRUE, recursively includes the virtual fields
    #   // in the virtual registers in the sub-blocks.
    #   //
    #   extern virtual function void get_virtual_fields (ref uvm_vreg_field fields[$],
    #                                                 input uvm_hier_e hier=UVM_HIER)


    #   // Function: get_block_by_name
    #   //
    #   // Finds a sub-block with the specified simple name.
    #   //
    #   // The name is the simple name of the block, not a hierarchical name.
    #   // relative to this block.
    #   // If no block with that name is found in this block, the sub-blocks
    #   // are searched for a block of that name and the first one to be found
    #   // is returned.
    #   //
    #   // If no blocks are found, returns ~None~.
    #   //
    #   extern virtual function uvm_reg_block get_block_by_name (string name)

    #
    #
    #   // Function: get_map_by_name
    #   //
    #   // Finds an address map with the specified simple name.
    #   //
    #   // The name is the simple name of the address map, not a hierarchical name.
    #   // relative to this block.
    #   // If no map with that name is found in this block, the sub-blocks
    #   // are searched for a map of that name and the first one to be found
    #   // is returned.
    #   //
    #   // If no address maps are found, returns ~None~.
    #   //
    #   extern virtual function UVMRegMap get_map_by_name (string name)
    def get_map_by_name(self, name) -> Optional[UVMRegMap]:
        maps: List[UVMRegMap] = []
        self.get_maps(maps)

        for i in range(len(maps)):
            if (maps[i].get_name() == name):
                return maps[i]

        for i in range(len(maps)):
            submaps: List[UVMRegMap] = []
            maps[i].get_submaps(submaps, UVM_HIER)

            for j in range(len(submaps)):
                if (submaps[j].get_name() == name):
                    return submaps[j]

        uvm_warning("RegModel", "Map with name '" + name + "' does not exist in block")
        return None



    #   // Function: get_reg_by_name
    #   //
    #   // Finds a register with the specified simple name.
    #   //
    #   // The name is the simple name of the register, not a hierarchical name.
    #   // relative to this block.
    #   // If no register with that name is found in this block, the sub-blocks
    #   // are searched for a register of that name and the first one to be found
    #   // is returned.
    #   //
    #   // If no registers are found, returns ~None~.
    #   //
    #   extern virtual function uvm_reg get_reg_by_name (string name)
    def get_reg_by_name(self, name) -> Optional[UVMReg]:

        for rg in self.regs.key_list():
            if rg.get_name() == name:
                return rg

        for blk_ in self.blks.key_list():
            # blk = blk_  # uvm_reg_block
            subregs = []  # uvm_reg[$]
            blk_.get_registers(subregs, UVM_HIER)

            for j in range(len(subregs)):
                if subregs[j].get_name() == name:
                    return subregs[j]

        uvm_warning("RegModel", "Unable to locate register '" + name +
                        "' in block '" + self.get_full_name() + "'")
        return None
        #endfunction: get_reg_by_name


    #   // Function: get_field_by_name
    #   //
    #   // Finds a field with the specified simple name.
    #   //
    #   // The name is the simple name of the field, not a hierarchical name.
    #   // relative to this block.
    #   // If no field with that name is found in this block, the sub-blocks
    #   // are searched for a field of that name and the first one to be found
    #   // is returned.
    #   //
    #   // If no fields are found, returns ~None~.
    #   //
    #   extern virtual function uvm_reg_field get_field_by_name (string name)
    #
    #
    #   // Function: get_mem_by_name
    #   //
    #   // Finds a memory with the specified simple name.
    #   //
    #   // The name is the simple name of the memory, not a hierarchical name.
    #   // relative to this block.
    #   // If no memory with that name is found in this block, the sub-blocks
    #   // are searched for a memory of that name and the first one to be found
    #   // is returned.
    #   //
    #   // If no memories are found, returns ~None~.
    #   //
    #   extern virtual function uvm_mem get_mem_by_name (string name)
    #
    #
    #   // Function: get_vreg_by_name
    #   //
    #   // Finds a virtual register with the specified simple name.
    #   //
    #   // The name is the simple name of the virtual register,
    #   // not a hierarchical name.
    #   // relative to this block.
    #   // If no virtual register with that name is found in this block,
    #   // the sub-blocks are searched for a virtual register of that name
    #   // and the first one to be found is returned.
    #   //
    #   // If no virtual registers are found, returns ~None~.
    #   //
    #   extern virtual function uvm_vreg get_vreg_by_name (string name)
    #
    #
    #   // Function: get_vfield_by_name
    #   //
    #   // Finds a virtual field with the specified simple name.
    #   //
    #   // The name is the simple name of the virtual field,
    #   // not a hierarchical name.
    #   // relative to this block.
    #   // If no virtual field with that name is found in this block,
    #   // the sub-blocks are searched for a virtual field of that name
    #   // and the first one to be found is returned.
    #   //
    #   // If no virtual fields are found, returns ~None~.
    #   //
    #   extern virtual function uvm_vreg_field get_vfield_by_name (string name)
    #

    #   //----------------
    #   // Group: Coverage
    #   //----------------


    #   // Function: build_coverage
    #   //
    #   // Check if all of the specified coverage model must be built.
    #   //
    #   // Check which of the specified coverage model must be built
    #   // in this instance of the block abstraction class,
    #   // as specified by calls to <uvm_reg::include_coverage()>.
    #   //
    #   // Models are specified by adding the symbolic value of individual
    #   // coverage model as defined in <uvm_coverage_model_e>.
    #   // Returns the sum of all coverage models to be built in the
    #   // block model.
    #   //
    #   extern protected function uvm_reg_cvr_t build_coverage(uvm_reg_cvr_t models)
    def build_coverage(self, models):
        build_coverage = UVM_NO_COVERAGE
        cov_arr = []
        uvm_reg_cvr_rsrc_db.read_by_name("uvm_reg::" + self.get_full_name(),
                "include_coverage",
                cov_arr, self)
        build_coverage = cov_arr[0]
        return build_coverage & models


    #   // Function: add_coverage
    #   //
    #   // Specify that additional coverage models are available.
    #   //
    #   // Add the specified coverage model to the coverage models
    #   // available in this class.
    #   // Models are specified by adding the symbolic value of individual
    #   // coverage model as defined in <uvm_coverage_model_e>.
    #   //
    #   // This method shall be called only in the constructor of
    #   // subsequently derived classes.
    #   //
    #   extern virtual protected function void add_coverage(uvm_reg_cvr_t models)
    def add_coverage(self, models):
        self.has_cover |= models


    #   // Function: has_coverage
    #   //
    #   // Check if block has coverage model(s)
    #   //
    #   // Returns TRUE if the block abstraction class contains a coverage model
    #   // for all of the models specified.
    #   // Models are specified by adding the symbolic value of individual
    #   // coverage model as defined in <uvm_coverage_model_e>.
    #   //
    #   extern virtual function bit has_coverage(uvm_reg_cvr_t models)
    def has_coverage(self, models):
        return ((self.has_cover & models) == models)


    #   // Function: set_coverage
    #   //
    #   // Turns on coverage measurement.
    #   //
    #   // Turns the collection of functional coverage measurements on or off
    #   // for this block and all blocks, registers, fields and memories within it.
    #   // The functional coverage measurement is turned on for every
    #   // coverage model specified using <uvm_coverage_model_e> symbolic
    #   // identifiers.
    #   // Multiple functional coverage models can be specified by adding
    #   // the functional coverage model identifiers.
    #   // All other functional coverage models are turned off.
    #   // Returns the sum of all functional
    #   // coverage models whose measurements were previously on.
    #   //
    #   // This method can only control the measurement of functional
    #   // coverage models that are present in the various abstraction classes,
    #   // then enabled during construction.
    #   // See the <uvm_reg_block::has_coverage()> method to identify
    #   // the available functional coverage models.
    #   //
    #   extern virtual function uvm_reg_cvr_t set_coverage(uvm_reg_cvr_t is_on)
    def set_coverage(self, is_on):
        self.cover_on = self.has_cover & is_on

        for rg_ in self.regs.key_list():
            rg = rg_
            rg.set_coverage(is_on)

        for mem in self.mems.key_list():
            mem.set_coverage(is_on)

        for blk_ in self.blks.key_list():
            blk = blk_
            blk.set_coverage(is_on)

        return self.cover_on


    #   // Function: get_coverage
    #   //
    #   // Check if coverage measurement is on.
    #   //
    #   // Returns TRUE if measurement for all of the specified functional
    #   // coverage models are currently on.
    #   // Multiple functional coverage models can be specified by adding the
    #   // functional coverage model identifiers.
    #   //
    #   // See <uvm_reg_block::set_coverage()> for more details.
    #   //
    #   extern virtual function bit get_coverage(uvm_reg_cvr_t is_on = UVM_CVR_ALL)
    def get_coverage(self, is_on=UVM_CVR_ALL):
        if self.has_coverage(is_on) == 0:
            return 0
        return ((self.cover_on & is_on) == is_on)


    #   // Function: sample
    #   //
    #   // Functional coverage measurement method
    #   //
    #   // This method is invoked by the block abstraction class
    #   // whenever an address within one of its address map
    #   // is successfully read or written.
    #   // The specified offset is the offset within the block,
    #   // not an absolute address.
    #   //
    #   // Empty by default, this method may be extended by the
    #   // abstraction class generator to perform the required sampling
    #   // in any provided functional coverage model.
    #   //
    #   protected virtual function void  sample(uvm_reg_addr_t offset,
    #                                           bit            is_read,
    #                                           UVMRegMap    map)
    #   endfunction
    def sample(self, offset, is_read, _map):
        pass

    #   // Function: sample_values
    #   //
    #   // Functional coverage measurement method for field values
    #   //
    #   // This method is invoked by the user
    #   // or by the <uvm_reg_block::sample_values()> method of the parent block
    #   // to trigger the sampling
    #   // of the current field values in the
    #   // block-level functional coverage model.
    #   // It recursively invokes the <uvm_reg_block::sample_values()>
    #   // and <uvm_reg::sample_values()> methods
    #   // in the blocks and registers in this block.
    #   //
    #   // This method may be extended by the
    #   // abstraction class generator to perform the required sampling
    #   // in any provided field-value functional coverage model.
    #   // If this method is extended, it MUST call super.sample_values().
    #   //
    #   extern virtual function void sample_values()
    def sample_values(self):
        for rg_ in self.regs.key_list():
            rg = rg_
            rg.sample_values()

        for blk_ in self.blks.key_list():
            blk = blk_
            blk.sample_values()


    #   /*local*/ extern function void XsampleX(uvm_reg_addr_t addr,
    #                                           bit            is_read,
    #                                           UVMRegMap    map)
    def XsampleX(self, addr, is_read, _map):
        self.sample(addr, is_read, _map)
        if (self.parent is not None):
            pass
            # ToDo: Call XsampleX in the parent block
            # with the offset and map within that block's context


    #   //--------------
    #   // Group: Access
    #   //--------------

    #   // Function: get_default_path
    #   //
    #   // Default access path
    #   //
    #   // Returns the default access path for this block.
    #   //
    #   extern virtual function uvm_path_e get_default_path()
    def get_default_path(self):
        if (self.default_path != UVM_DEFAULT_PATH):
            return self.default_path
        if (self.parent is not None):
            return self.parent.get_default_path()
        return UVM_FRONTDOOR


    #   // Function: reset
    #   //
    #   // Reset the mirror for this block.
    #   //
    #   // Sets the mirror value of all registers in the block and sub-blocks
    #   // to the reset value corresponding to the specified reset event.
    #   // See <uvm_reg_field::reset()> for more details.
    #   // Does not actually set the value of the registers in the design,
    #   // only the values mirrored in their corresponding mirror.
    #   //
    #   extern virtual function void reset(string kind = "HARD")
    def reset(self, kind="HARD"):
        #
        for rg_ in self.regs.key_list():
            rg = rg_
            rg.reset(kind)

        for blk_ in self.blks.key_list():
            blk = blk_  # uvm_reg_block
            blk.reset(kind)
        #endfunction


    #   // Function: needs_update
    #   //
    #   // Check if DUT registers need to be written
    #   //
    #   // If a mirror value has been modified in the abstraction model
    #   // without actually updating the actual register
    #   // (either through randomization or via the <uvm_reg::set()> method,
    #   // the mirror and state of the registers are outdated.
    #   // The corresponding registers in the DUT need to be updated.
    #   //
    #   // This method returns TRUE if the state of at least one register in
    #   // the block or sub-blocks needs to be updated to match the mirrored
    #   // values.
    #   // The mirror values, or actual content of registers, are not modified.
    #   // For additional information, see <uvm_reg_block::update()> method.
    #   //
    #   extern virtual function bit needs_update()
    def needs_update(self):

        for rg_ in self.regs.key_list():
            rg = rg_
            if rg.needs_update():
                return 1
        for blk_ in self.blks.key_list():
            blk = blk_
            if blk.needs_update():
                return 1
        return 0

    #   // Task: update
    #   //
    #   // Batch update of register.
    #   //
    #   // Using the minimum number of write operations, updates the registers
    #   // in the design to match the mirrored values in this block and sub-blocks.
    #   // The update can be performed using the physical
    #   // interfaces (front-door access) or back-door accesses.
    #   // This method performs the reverse operation of <uvm_reg_block::mirror()>.
    #   //
    #   extern virtual task update(output uvm_status_e       status,
    #                              input  uvm_path_e         path = UVM_DEFAULT_PATH,
    #                              input  uvm_sequence_base  parent = None,
    #                              input  int                prior = -1,
    #                              input  uvm_object         extension = None,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)
    async def update(self, status, path=UVM_DEFAULT_PATH, parent=None, prior=-1, extension=None, fname="",
            lineno=0):
        uvm_check_output_args([status])
        stat_all = UVM_IS_OK

        if self.needs_update() is False:
            uvm_info("RegModel", sv.sformatf("%s:%0d - RegModel block %s does not need updating",
                         fname, lineno, self.get_name()), UVM_HIGH)
            return


        uvm_info("RegModel", sv.sformatf("%s:%0d - Updating model block %s with %s path",
            fname, lineno, self.get_name(), path), UVM_HIGH)

        for rg_ in self.regs.key_list():
            rg = rg_
            if rg.needs_update():
                stat = []
                await rg.update(stat, path, None, parent, prior, extension)
                if (stat[0] != UVM_IS_OK and stat[0] != UVM_HAS_X):
                    uvm_error("RegModel", sv.sformatf("Register \"%s\" could not be updated",
                        rg.get_full_name()))
                    status.append(stat[0])
                    return

        for blk_ in self.blks.key_list():
            blk = blk_
            stat_blk = []
            await blk.update(status,path,parent,prior,extension,fname,lineno)
            if (stat_blk[0] != UVM_IS_OK and stat_blk[0] != UVM_HAS_X):
                stat_all = stat_blk
        status.append(stat_all)


    #   // Task: mirror
    #   //
    #   // Update the mirrored values
    #   //
    #   // Read all of the registers in this block and sub-blocks and update their
    #   // mirror values to match their corresponding values in the design.
    #   // The mirroring can be performed using the physical interfaces
    #   // (front-door access) or back-door accesses.
    #   // If the ~check~ argument is specified as <UVM_CHECK>,
    #   // an error message is issued if the current mirrored value
    #   // does not match the actual value in the design.
    #   // This method performs the reverse operation of <uvm_reg_block::update()>.
    #   //
    #   extern virtual task mirror(output uvm_status_e       status,
    #                              input  uvm_check_e        check = UVM_NO_CHECK,
    #                              input  uvm_path_e         path  = UVM_DEFAULT_PATH,
    #                              input  uvm_sequence_base  parent = None,
    #                              input  int                prior = -1,
    #                              input  uvm_object         extension = None,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)
    async def mirror(self, status, check=UVM_NO_CHECK, path=UVM_DEFAULT_PATH, parent=None, prior=-1,
            extension=None, fname="", lineno=0):
        final_status = UVM_IS_OK

        for rg_ in self.regs.key_list():
            rg = rg_
            curr_stat = []
            await rg.mirror(curr_stat, check, path, None, parent, prior, extension, fname, lineno)
            if (curr_stat[0] != UVM_IS_OK and curr_stat[0] != UVM_HAS_X):
                final_status = curr_stat[0]


        for blk_ in self.blks.key_list():
            blk = blk_  # uvm_reg_block

            curr_stat = []
            await blk.mirror(curr_stat, check, path, parent, prior, extension, fname, lineno)
            if (curr_stat[0] != UVM_IS_OK and curr_stat[0] != UVM_HAS_X):
                final_status = curr_stat[0]
        status.append(final_status)


    #   // Task: write_reg_by_name
    #   //
    #   // Write the named register
    #   //
    #   // Equivalent to <get_reg_by_name()> followed by <uvm_reg::write()>
    #   //
    #   extern virtual task write_reg_by_name(
    #                              output uvm_status_e        status,
    #                              input  string              name,
    #                              input  uvm_reg_data_t      data,
    #                              input  uvm_path_e     path = UVM_DEFAULT_PATH,
    #                              input  UVMRegMap         map = None,
    #                              input  uvm_sequence_base   parent = None,
    #                              input  int                 prior = -1,
    #                              input  uvm_object          extension = None,
    #                              input  string              fname = "",
    #                              input  int                 lineno = 0)
    async def write_reg_by_name(self, status, name, data, path=UVM_DEFAULT_PATH,
            map=None, parent=None, prior=-1, extension=None, fname="",
            lineno=0):
        self.fname = fname
        self.lineno = lineno

        status = UVM_NOT_OK
        rg = self.get_reg_by_name(name)
        if rg is not None:
            await rg.write(status, data, path, map, parent, prior, extension)


    #   // Task: read_reg_by_name
    #   //
    #   // Read the named register
    #   //
    #   // Equivalent to <get_reg_by_name()> followed by <uvm_reg::read()>
    #   //
    #   extern virtual task read_reg_by_name(
    #                              output uvm_status_e       status,
    #                              input  string             name,
    #                              output uvm_reg_data_t     data,
    #                              input  uvm_path_e    path = UVM_DEFAULT_PATH,
    #                              input  UVMRegMap        map = None,
    #                              input  uvm_sequence_base  parent = None,
    #                              input  int                prior = -1,
    #                              input  uvm_object         extension = None,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)


    #   // Task: write_mem_by_name
    #   //
    #   // Write the named memory
    #   //
    #   // Equivalent to <get_mem_by_name()> followed by <uvm_mem::write()>
    #   //
    #   extern virtual task write_mem_by_name(
    #                              output uvm_status_e       status,
    #                              input  string             name,
    #                              input  uvm_reg_addr_t     offset,
    #                              input  uvm_reg_data_t     data,
    #                              input  uvm_path_e    path = UVM_DEFAULT_PATH,
    #                              input  UVMRegMap        map = None,
    #                              input  uvm_sequence_base  parent = None,
    #                              input  int                prior = -1,
    #                              input  uvm_object         extension = None,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)


    #   // Task: read_mem_by_name
    #   //
    #   // Read the named memory
    #   //
    #   // Equivalent to <get_mem_by_name()> followed by <uvm_mem::read()>
    #   //
    #   extern virtual task read_mem_by_name(
    #                              output uvm_status_e       status,
    #                              input  string             name,
    #                              input  uvm_reg_addr_t     offset,
    #                              output uvm_reg_data_t     data,
    #                              input  uvm_path_e    path = UVM_DEFAULT_PATH,
    #                              input  UVMRegMap        map = None,
    #                              input  uvm_sequence_base  parent = None,
    #                              input  int                prior = -1,
    #                              input  uvm_object         extension = None,
    #                              input  string             fname = "",
    #                              input  int                lineno = 0)


    #   extern virtual task readmemh(string filename)

    #   extern virtual task writememh(string filename)


    #   //----------------
    #   // Group: Backdoor
    #   //----------------

    #   // Function: get_backdoor
    #   //
    #   // Get the user-defined backdoor for all registers in this block
    #   //
    #   // Return the user-defined backdoor for all register in this
    #   // block and all sub-blocks -- unless overridden by a backdoor set
    #   // in a lower-level block or in the register itself.
    #   //
    #   // If ~inherited~ is TRUE, returns the backdoor of the parent block
    #   // if none have been specified for this block.
    #   //
    #   extern function uvm_reg_backdoor get_backdoor(bit inherited = 1)
    def get_backdoor(self, inherited=True):
        if (self.backdoor is None and inherited):
            blk = self.get_parent()
            while (blk is not None):
                bkdr = blk.get_backdoor()
                if (bkdr is not None):
                    return bkdr
                blk = blk.get_parent()
        return self.backdoor
        #endfunction: get_backdoor


    #   // Function: set_backdoor
    #   //
    #   // Set the user-defined backdoor for all registers in this block
    #   //
    #   // Defines the backdoor mechanism for all registers instantiated
    #   // in this block and sub-blocks, unless overridden by a definition
    #   // in a lower-level block or register.
    #   //
    #   extern function void set_backdoor (uvm_reg_backdoor bkdr,
    #                                      string fname = "",
    #                                      int lineno = 0)


    #   // Function:  clear_hdl_path
    #   //
    #   // Delete HDL paths
    #   //
    #   // Remove any previously specified HDL path to the block instance
    #   // for the specified design abstraction.
    #   //
    #   extern function void clear_hdl_path (string kind = "RTL")
    #
    #

    #   // Function:  add_hdl_path
    #   //
    #   // Add an HDL path
    #   //
    #   // Add the specified HDL path to the block instance for the specified
    #   // design abstraction. This method may be called more than once for the
    #   // same design abstraction if the block is physically duplicated
    #   // in the design abstraction
    #   //
    #   extern function void add_hdl_path (string path, string kind = "RTL")
    def add_hdl_path(self, path, kind="RTL"):
        #  uvm_queue #(string) paths
        paths = self.hdl_paths_pool.get(kind)
        paths.push_back(path)


    #   // Function:   has_hdl_path
    #   //
    #   // Check if a HDL path is specified
    #   //
    #   // Returns TRUE if the block instance has a HDL path defined for the
    #   // specified design abstraction. If no design abstraction is specified,
    #   // uses the default design abstraction specified for this block or
    #   // the nearest block ancestor with a specified default design abstraction.
    #   //
    def has_hdl_path(self, kind=""):
        if kind == "":
            kind = self.get_default_hdl_path()
        return self.hdl_paths_pool.exists(kind)


    #   // Function:  get_hdl_path
    #   //
    #   // Get the incremental HDL path(s)
    #   //
    #   // Returns the HDL path(s) defined for the specified design abstraction
    #   // in the block instance.
    #   // Returns only the component of the HDL paths that corresponds to
    #   // the block, not a full hierarchical path
    #   //
    #   // If no design abstraction is specified, the default design abstraction
    #   // for this block is used.
    #   //
    #   extern function void get_hdl_path (ref string paths[$], input string kind = "")


    #   // Function:  get_full_hdl_path
    #   //
    #   // Get the full hierarchical HDL path(s)
    #   //
    #   // Returns the full hierarchical HDL path(s) defined for the specified
    #   // design abstraction in the block instance.
    #   // There may be more than one path returned even
    #   // if only one path was defined for the block instance, if any of the
    #   // parent components have more than one path defined for the same design
    #   // abstraction
    #   //
    #   // If no design abstraction is specified, the default design abstraction
    #   // for each ancestor block is used to get each incremental path.
    #   //
    #   extern function void get_full_hdl_path (ref string paths[$],
    #                                           input string kind = "",
    #                                           string separator = ".")
    def get_full_hdl_path(self, paths, kind="", separator="."):

        if kind == "":
            kind = self.get_default_hdl_path()

        paths.clear()
        if self.is_hdl_path_root(kind):
            if self.root_hdl_paths[kind] != "":
                paths.append(self.root_hdl_paths[kind])
            return


        if not self.has_hdl_path(kind):
            uvm_error("RegModel", "Block " + self.get_full_name() +
                " does not have hdl path defined for abstraction '" + kind + "'")
            return

        hdl_paths = self.hdl_paths_pool.get(kind)  # uvm_queue #(string)
        parent_paths = []

        if self.parent is not None:
            self.parent.get_full_hdl_path(parent_paths, kind, separator)

        for i in range(len(hdl_paths)):
            hdl_path = hdl_paths.get(i)

            if len(parent_paths) == 0:
                if hdl_path != "":
                    paths.append(hdl_path)
                continue

            for j in range(len(parent_paths)):
                if hdl_path == "":
                    paths.append(parent_paths[j])
                else:
                    paths.append(parent_paths[j] + separator + hdl_path)


    #   // Function: set_default_hdl_path
    #   //
    #   // Set the default design abstraction
    #   //
    #   // Set the default design abstraction for this block instance.
    #   //
    #   extern function void   set_default_hdl_path (string kind)
    def set_default_hdl_path(self, kind: str):
        if kind == "":
            if self.parent is None:
                uvm_error("RegModel", ("Block has no parent. " +
                   "Must specify a valid HDL abstraction (kind)"))
            else:
                kind = self.parent.get_default_hdl_path()
        self.default_hdl_path = kind


    #   // Function:  get_default_hdl_path
    #   //
    #   // Get the default design abstraction
    #   //
    #   // Returns the default design abstraction for this block instance.
    #   // If a default design abstraction has not been explicitly set for this
    #   // block instance, returns the default design abstraction for the
    #   // nearest block ancestor.
    #   // Returns "" if no default design abstraction has been specified.
    #   //
    #   extern function string get_default_hdl_path ()
    def get_default_hdl_path(self) -> str:
        if self.default_hdl_path == "" and self.parent is not None:
            return self.parent.get_default_hdl_path()
        return self.default_hdl_path


    #   // Function: set_hdl_path_root
    #   //
    #   // Specify a root HDL path
    #   //
    #   // Set the specified path as the absolute HDL path to the block instance
    #   // for the specified design abstraction.
    #   // This absolute root path is prepended to all hierarchical paths
    #   // under this block. The HDL path of any ancestor block is ignored.
    #   // This method overrides any incremental path for the
    #   // same design abstraction specified using <add_hdl_path>.
    #   //
    #   extern function void set_hdl_path_root (string path, string kind = "RTL")
    # set_hdl_path_root
    def set_hdl_path_root(self, path, kind="RTL"):
        if kind == "":
            kind = self.get_default_hdl_path()
        self.root_hdl_paths[kind] = path

    #   // Function: is_hdl_path_root
    #   //
    #   // Check if this block has an absolute path
    #   //
    #   // Returns TRUE if an absolute HDL path to the block instance
    #   // for the specified design abstraction has been defined.
    #   // If no design abstraction is specified, the default design abstraction
    #   // for this block is used.
    #   //
    #   extern function bit is_hdl_path_root (string kind = "")
    def is_hdl_path_root(self, kind=""):
        if kind == "":
            kind = self.get_default_hdl_path()
        return self.root_hdl_paths.exists(kind)
    #endfunction



    #   extern virtual function void   do_print      (uvm_printer printer)
    #   extern virtual function void   do_copy       (uvm_object rhs)
    #   extern virtual function bit    do_compare    (uvm_object  rhs,
    #                                                 uvm_comparer comparer)
    #   extern virtual function void   do_pack       (uvm_packer packer)
    #   extern virtual function void   do_unpack     (uvm_packer packer)

    #   extern virtual function string convert2string ()

    #   extern virtual function uvm_object clone()

    #   extern local function void Xinit_address_mapsX()
    def Xinit_address_mapsX(self):
        for map_ in self.maps.key_list():
            map_.Xinit_address_mapX()
        # map.Xverify_map_configX()



#------------------------------------------------------------------------
#
#---------------
# Initialization
#---------------
#
#
#
#
#
#--------------------------
# Get Hierarchical Elements
#--------------------------
#
# get_fields
#
#function void uvm_reg_block::get_fields(ref uvm_reg_field fields[$],
#                                        input uvm_hier_e hier=UVM_HIER)
#
#   foreach (regs[rg_]):
#     uvm_reg rg = rg_
#     rg.get_fields(fields)
#   end
#
#   if (hier == UVM_HIER)
#     foreach (blks[blk_])
#     begin
#       uvm_reg_block blk = blk_
#       blk.get_fields(fields)
#     end
#
#endfunction: get_fields
#
#
# get_virtual_fields
#
#function void uvm_reg_block::get_virtual_fields(ref uvm_vreg_field fields[$],
#                                                input uvm_hier_e hier=UVM_HIER)
#
#   foreach (vregs[vreg_]):
#     uvm_vreg vreg = vreg_
#     vreg.get_fields(fields)
#   end
#
#   if (hier == UVM_HIER)
#     foreach (blks[blk_]):
#       uvm_reg_block blk = blk_
#       blk.get_virtual_fields(fields)
#     end
#endfunction: get_virtual_fields
#
#
#
#
# get_virtual_registers
#
#function void uvm_reg_block::get_virtual_registers(ref uvm_vreg regs[$],
#                                                   input uvm_hier_e hier=UVM_HIER)
#
#   foreach (vregs[rg])
#     regs.push_back(rg)
#
#   if (hier == UVM_HIER)
#     foreach (blks[blk_]):
#       uvm_reg_block blk = blk_
#       blk.get_virtual_registers(regs)
#     end
#endfunction: get_virtual_registers
#
#
#
#
#
#
#
#
# find_blocks
#
#function int uvm_reg_block::find_blocks(input string        name,
#                                        ref   uvm_reg_block blks[$],
#                                        input uvm_reg_block root = None,
#                                        input uvm_object    accessor = None)
#
#   uvm_resource_pool rpl = uvm_resource_pool::get()
#   uvm_resource_types::rsrc_q_t rs
#
#   blks.delete()
#
#   if (root is not None) name = {root.get_full_name(), ".", name}
#
#   rs = rpl.lookup_regex(name, "uvm_reg::")
#   for (int i = 0; i < rs.size(); i++):
#      uvm_resource#(uvm_reg_block) blk
#      if (!$cast(blk, rs.get(i))) continue
#      blks.push_back(blk.read(accessor))
#   end
#
#   return blks.size()
#endfunction
#
#
# find_blocks
#
#function uvm_reg_block uvm_reg_block::find_block(input string        name,
#                                                 input uvm_reg_block root = None,
#                                                 input uvm_object    accessor = None)
#
#   uvm_reg_block blks[$]
#   if (!find_blocks(name, blks, root, accessor))
#      return None
#
#   if (blks.size() > 1):
#      `uvm_warning("MRTH1BLK",
#                   {"More than one block matched the name \"", name, "\"."})
#   end
#
#
#   return blks[0]
#endfunction
#
#
#
#------------
# Get-By-Name
#------------
#
# get_block_by_name
#
#function uvm_reg_block uvm_reg_block::get_block_by_name(string name)
#
#   if (get_name() == name)
#     return this
#
#   foreach (blks[blk_]):
#     uvm_reg_block blk = blk_
#
#     if (blk.get_name() == name)
#       return blk
#   end
#
#   foreach (blks[blk_]):
#      uvm_reg_block blk = blk_
#      uvm_reg_block subblks[$]
#      blk_.get_blocks(subblks, UVM_HIER)
#
#      foreach (subblks[j])
#         if (subblks[j].get_name() == name)
#            return subblks[j]
#   end
#
#   `uvm_warning("RegModel", {"Unable to locate block '",name,
#                "' in block '",get_full_name(),"'"})
#   return None
#
#endfunction: get_block_by_name
#
#
#
#
# get_vreg_by_name
#
#function uvm_vreg uvm_reg_block::get_vreg_by_name(string name)
#
#   foreach (vregs[rg_]):
#     uvm_vreg rg = rg_
#     if (rg.get_name() == name)
#       return rg
#   end
#
#   foreach (blks[blk_]):
#      uvm_reg_block blk = blk_
#      uvm_vreg subvregs[$]
#      blk_.get_virtual_registers(subvregs, UVM_HIER)
#
#      foreach (subvregs[j])
#         if (subvregs[j].get_name() == name)
#            return subvregs[j]
#   end
#
#   `uvm_warning("RegModel", {"Unable to locate virtual register '",name,
#                "' in block '",get_full_name(),"'"})
#   return None
#
#endfunction: get_vreg_by_name
#
#
# get_mem_by_name
#
#function uvm_mem uvm_reg_block::get_mem_by_name(string name)
#
#   foreach (mems[mem_]):
#     uvm_mem mem = mem_
#     if (mem.get_name() == name)
#       return mem
#   end
#
#   foreach (blks[blk_]):
#      uvm_reg_block blk = blk_
#      uvm_mem submems[$]
#      blk_.get_memories(submems, UVM_HIER)
#
#      foreach (submems[j])
#         if (submems[j].get_name() == name)
#            return submems[j]
#   end
#
#   `uvm_warning("RegModel", {"Unable to locate memory '",name,
#                "' in block '",get_full_name(),"'"})
#   return None
#
#endfunction: get_mem_by_name
#
#
# get_field_by_name
#
#function uvm_reg_field uvm_reg_block::get_field_by_name(string name)
#
#   foreach (regs[rg_]):
#      uvm_reg rg = rg_
#      uvm_reg_field fields[$]
#
#      rg.get_fields(fields)
#      foreach (fields[i])
#        if (fields[i].get_name() == name)
#          return fields[i]
#   end
#
#   foreach (blks[blk_]):
#      uvm_reg_block blk = blk_
#      uvm_reg subregs[$]
#      blk_.get_registers(subregs, UVM_HIER)
#
#      foreach (subregs[j]):
#         uvm_reg_field fields[$]
#         subregs[j].get_fields(fields)
#         foreach (fields[i])
#            if (fields[i].get_name() == name)
#               return fields[i]
#      end
#   end
#
#   `uvm_warning("RegModel", {"Unable to locate field '",name,
#                "' in block '",get_full_name(),"'"})
#
#   return None
#
#endfunction: get_field_by_name
#
#
# get_vfield_by_name
#
#function uvm_vreg_field uvm_reg_block::get_vfield_by_name(string name)
#
#   foreach (vregs[rg_]):
#      uvm_vreg rg =rg_
#      uvm_vreg_field fields[$]
#
#      rg.get_fields(fields)
#      foreach (fields[i])
#        if (fields[i].get_name() == name)
#          return fields[i]
#   end
#
#   foreach (blks[blk_]):
#      uvm_reg_block blk = blk_
#      uvm_vreg subvregs[$]
#      blk_.get_virtual_registers(subvregs, UVM_HIER)
#
#      foreach (subvregs[j]):
#         uvm_vreg_field fields[$]
#         subvregs[j].get_fields(fields)
#         foreach (fields[i])
#            if (fields[i].get_name() == name)
#               return fields[i]
#      end
#   end
#
#   `uvm_warning("RegModel", {"Unable to locate virtual field '",name,
#                "' in block '",get_full_name(),"'"})
#
#   return None
#
#endfunction: get_vfield_by_name
#
#
#----------------
# Run-Time Access
#----------------
#
#
#
# read_reg_by_name
#
#task uvm_reg_block::read_reg_by_name(output uvm_status_e  status,
#                                     input  string             name,
#                                     output uvm_reg_data_t     data,
#                                     input  uvm_path_e    path = UVM_DEFAULT_PATH,
#                                     input  UVMRegMap     map = None,
#                                     input  uvm_sequence_base  parent = None,
#                                     input  int                prior = -1,
#                                     input  uvm_object         extension = None,
#                                     input  string             fname = "",
#                                     input  int                lineno = 0)
#   uvm_reg rg
#   self.fname = fname
#   self.lineno = lineno
#
#   status = UVM_NOT_OK
#   rg = self.get_reg_by_name(name)
#   if (rg is not None)
#     rg.read(status, data, path, map, parent, prior, extension)
#endtask: read_reg_by_name
#
#
# write_mem_by_name
#
#task uvm_reg_block::write_mem_by_name(output uvm_status_e  status,
#                                          input  string             name,
#                                          input  uvm_reg_addr_t     offset,
#                                          input  uvm_reg_data_t     data,
#                                          input  uvm_path_e    path = UVM_DEFAULT_PATH,
#                                          input  UVMRegMap     map = None,
#                                          input  uvm_sequence_base  parent = None,
#                                          input  int                prior = -1,
#                                          input  uvm_object         extension = None,
#                                          input  string             fname = "",
#                                          input  int                lineno = 0)
#   uvm_mem mem
#   self.fname = fname
#   self.lineno = lineno
#
#   status = UVM_NOT_OK
#   mem = get_mem_by_name(name)
#   if (mem is not None)
#     mem.write(status, offset, data, path, map, parent, prior, extension)
#endtask: write_mem_by_name
#
#
# read_mem_by_name
#
#task uvm_reg_block::read_mem_by_name(output uvm_status_e  status,
#                                         input  string             name,
#                                         input  uvm_reg_addr_t     offset,
#                                         output uvm_reg_data_t     data,
#                                         input  uvm_path_e    path = UVM_DEFAULT_PATH,
#                                         input  UVMRegMap     map = None,
#                                         input  uvm_sequence_base  parent = None,
#                                         input  int                prior = -1,
#                                         input  uvm_object         extension = None,
#                                         input  string             fname = "",
#                                         input  int                lineno = 0)
#   uvm_mem mem
#   self.fname = fname
#   self.lineno = lineno
#
#   status = UVM_NOT_OK
#   mem = get_mem_by_name(name)
#   if (mem is not None)
#     mem.read(status, offset, data, path, map, parent, prior, extension)
#endtask: read_mem_by_name
#
#
# readmemh
#
#task uvm_reg_block::readmemh(string filename)
#   // TODO
#endtask: readmemh
#
#
# writememh
#
#task uvm_reg_block::writememh(string filename)
#   // TODO
#endtask: writememh
#
#
#---------------
# Map Management
#---------------
#
#----------------
# Group- Backdoor
#----------------
#
# set_backdoor
#
#function void uvm_reg_block::set_backdoor(uvm_reg_backdoor bkdr,
#                                          string               fname = "",
#                                          int                  lineno = 0)
#   bkdr.fname = fname
#   bkdr.lineno = lineno
#   if (self.backdoor is not None &&
#       self.backdoor.has_update_threads()):
#      `uvm_warning("RegModel", "Previous register backdoor still has update threads running. Backdoors with active mirroring should only be set before simulation starts.")
#   end
#   self.backdoor = bkdr
#endfunction: set_backdoor
#
#
#
# clear_hdl_path
#
#function void uvm_reg_block::clear_hdl_path(string kind = "RTL")
#
#  if (kind == "ALL"):
#    hdl_paths_pool = new("hdl_paths")
#    return
#  end
#
#  if (kind == "")
#    kind = get_default_hdl_path()
#
#  if (!hdl_paths_pool.exists(kind)):
#    `uvm_warning("RegModel",{"Unknown HDL Abstraction '",kind,"'"})
#    return
#  end
#
#  hdl_paths_pool.delete(kind)
#endfunction
#
#
#
#
#
# get_hdl_path
#
#function void uvm_reg_block::get_hdl_path(ref string paths[$], input string kind = "")
#
#  uvm_queue #(string) hdl_paths
#
#  if (kind == "")
#    kind = get_default_hdl_path()
#
#  if (!has_hdl_path(kind)):
#    `uvm_error("RegModel",{"Block does not have hdl path defined for abstraction '",kind,"'"})
#    return
#  end
#
#  hdl_paths = hdl_paths_pool.get(kind)
#
#  for (int i=0; i<hdl_paths.size();i++)
#    paths.push_back(hdl_paths.get(i))
#
#endfunction
#
#
#
#
#
#
#
#
#
#
#
#
#----------------------------------
# Group- Basic Object Operations
#----------------------------------
#
# do_print
#function void uvm_reg_block::do_print (uvm_printer printer)
#  super.do_print(printer)
#
#  foreach(blks[i]):
#     uvm_reg_block b = i
#     uvm_object obj = b
#     printer.print_object(obj.get_name(), obj)
#  end
#
#  foreach(regs[i]):
#     uvm_reg r = i
#     uvm_object obj = r
#     printer.print_object(obj.get_name(), obj)
#  end
#
#  foreach(vregs[i]):
#     uvm_vreg r = i
#     uvm_object obj = r
#     printer.print_object(obj.get_name(), obj)
#  end
#
#  foreach(mems[i]):
#     uvm_mem m = i
#     uvm_object obj = m
#     printer.print_object(obj.get_name(), obj)
#  end
#
#  foreach(maps[i]):
#     UVMRegMap m = i
#     uvm_object obj = m
#     printer.print_object(obj.get_name(), obj)
#  end
#
#endfunction
#
#
#
# clone
#
#function uvm_object uvm_reg_block::clone()
#  `uvm_fatal("RegModel","RegModel blocks cannot be cloned")
#  return None
#endfunction
#
# do_copy
#
#function void uvm_reg_block::do_copy(uvm_object rhs)
#  `uvm_fatal("RegModel","RegModel blocks cannot be copied")
#endfunction
#
#
# do_compare
#
#function bit uvm_reg_block::do_compare (uvm_object  rhs,
#                                        uvm_comparer comparer)
#  `uvm_warning("RegModel","RegModel blocks cannot be compared")
#  return 0
#endfunction
#
#
# do_pack
#
#function void uvm_reg_block::do_pack (uvm_packer packer)
#  `uvm_warning("RegModel","RegModel blocks cannot be packed")
#endfunction
#
#
# do_unpack
#
#function void uvm_reg_block::do_unpack (uvm_packer packer)
#  `uvm_warning("RegModel","RegModel blocks cannot be unpacked")
#endfunction
#
#// convert2string
#
#def string uvm_reg_block::convert2string(self):
#   string image
#   string maps[]
#   string blk_maps[]
#   bit         single_map
#   uvm_endianness_e endian
#   string prefix = "  "
#
#`ifdef TODO
#   single_map = 1
#   if (map == ""):
#      self.get_maps(maps)
#      if (maps.size() > 1) single_map = 0
#   end
#
#   if (single_map):
#      $sformat(image, "%sBlock %s", prefix, self.get_full_name())
#
#      if (map != "")
#        $sformat(image, "%s.%s", image, map)
#
#      endian = self.get_endian(map)
#
#      $sformat(image, "%s -- %0d bytes (%s)", image,
#               self.get_n_bytes(map), endian.name())
#
#      foreach (blks[i]):
#         string img
#         img = blks[i].convert2string({prefix, "   "}, blk_maps[i])
#         image = {image, "\n", img}
#      end
#
#   end
#   else begin
#      $sformat(image, "%Block %s", prefix, self.get_full_name())
#      foreach (maps[i]):
#         string img
#         endian = self.get_endian(maps[i])
#         $sformat(img, "%s   Map \"%s\" -- %0d bytes (%s)",
#                  prefix, maps[i],
#                  self.get_n_bytes(maps[i]), endian.name())
#         image = {image, "\n", img}
#
#         self.get_blocks(blks, blk_maps, maps[i])
#         foreach (blks[j]):
#            img = blks[j].convert2string({prefix, "      "},
#                                    blk_maps[j])
#            image = {image, "\n", img}
#         end
#
#         self.get_subsys(sys, blk_maps, maps[i])
#         foreach (sys[j]):
#            img = sys[j].convert2string({prefix, "      "},
#                                   blk_maps[j])
#            image = {image, "\n", img}
#         end
#      end
#   end
#`endif
#   return image



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.base import sv
from uvm.comps import UVMScoreboard
from uvm.macros import uvm_component_utils, uvm_info
from uvm.tlm1 import UVMAnalysisImp
from ubus_transfer import *

#//------------------------------------------------------------------------------
#//
#// CLASS: ubus_example_scoreboard
#//
#//------------------------------------------------------------------------------
#class ubus_example_scoreboard extends uvm_scoreboard


class ubus_example_scoreboard(UVMScoreboard):
    #
    #  uvm_analysis_imp#(ubus_transfer, ubus_example_scoreboard) item_collected_export
    #
    #
    #  protected int unsigned m_mem_expected[int unsigned]
    #
    #  // Provide implementations of virtual methods such as get_type_name and create
    #  `uvm_component_utils_begin(ubus_example_scoreboard)
    #    `uvm_field_int(disable_scoreboard, UVM_DEFAULT)
    #    `uvm_field_int(num_writes, UVM_DEFAULT|UVM_DEC)
    #    `uvm_field_int(num_init_reads, UVM_DEFAULT|UVM_DEC)
    #    `uvm_field_int(num_uninit_reads, UVM_DEFAULT|UVM_DEC)
    #  `uvm_component_utils_end
    #

    def __init__(self, name, parent):
        UVMScoreboard.__init__(self, name, parent)
        self.disable_scoreboard = False
        self.num_writes = 0
        self.num_init_reads = 0
        self.num_uninit_reads = 0
        self.sbd_error = False
        self.m_mem_expected = {}

    #  //build_phase
    def build_phase(self, phase):
        self.item_collected_export = UVMAnalysisImp("item_collected_export", self)

    #  // write
    #  virtual function void write(ubus_transfer trans)
    def write(self, trans):
        if self.disable_scoreboard is False:
            self.memory_verify(trans)


    #  // memory_verify
    #  protected function void memory_verify(input ubus_transfer trans)
    def memory_verify(self, trans):
        data = 0
        exp = 0
        for i in range(trans.size):
            # Check to see if entry in associative array for this address when read
            # If so, check that transfer data matches associative array data.
            if (trans.addr + i) in self.m_mem_expected:
                if trans.read_write == READ:
                    data = trans.data[i]
                    uvm_info(self.get_type_name(),
                        sv.sformatf("%s to existing address...Checking address : %0h with data : %0h",
                            str(trans.read_write), int(trans.addr), int(data)), UVM_LOW)

                    if not (self.m_mem_expected[trans.addr + i] == trans.data[i]):
                        exp = self.m_mem_expected[trans.addr + i]
                        uvm_error(self.get_type_name(),
                          sv.sformatf("Read data mismatch.  Expected : %0h.  Actual : %0h",
                          exp, data))
                        self.sbd_error = True

                    self.num_init_reads += 1

                if trans.read_write == WRITE:
                    data = trans.data[i]
                    uvm_info(self.get_type_name(),
                      sv.sformatf("Op %s to existing address...Updating address : %0h with data : %0h",
                          str(trans.read_write), int(trans.addr + i), int(data)), UVM_LOW)
                    self.m_mem_expected[trans.addr + i] = trans.data[i]
                    self.num_writes += 1

            # Check to see if entry in associative array for this address
            # If not, update the location regardless if read or write.
            else:
                data = trans.data[i]
                uvm_info(self.get_type_name(),
                    sv.sformatf("%s to empty address...Updating address : %0h with data : %0h",
                    str(trans.read_write), int(trans.addr + i), int(data)), UVM_LOW)
                self.m_mem_expected[trans.addr + i] = trans.data[i]
                if trans.read_write == READ:
                    self.num_uninit_reads += 1
                elif (trans.read_write == WRITE):
                    self.num_writes += 1
        #  endfunction : memory_verify


    #  // report_phase
    #  virtual function void report_phase(uvm_phase phase)
    #    if(!disable_scoreboard):
    #      `uvm_info(get_type_name(),
    #        $sformatf("Reporting scoreboard information...\n%s", this.sprint()), UVM_LOW)
    #    end
    #  endfunction : report_phase

    #endclass : ubus_example_scoreboard
uvm_component_utils(ubus_example_scoreboard)



#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2010 Cadence Design Systems, Inc.
#//   Copyright 2010 Synopsys, Inc.
#//   Copyright 2019 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

from uvm.seq import UVMSequenceItem
from uvm.macros import *
from uvm.base.sv import sv

#//----------------------------------------------------------------------------
#//
#// ubus transfer enums, parameters, and events
#//
#//----------------------------------------------------------------------------

# ubus_read_write_enum;
NOP = 0
READ = 1
WRITE = 2

#//----------------------------------------------------------------------------
#//
#// CLASS: ubus_transfer
#//
#//----------------------------------------------------------------------------


class ubus_transfer(UVMSequenceItem):


    # TODO add these constraints
    #  constraint c_data_wait_size {
    #    data.size() == size;
    #    wait_state.size() == size;
    #  }

    def __init__(self, name="ubus_transfer_inst"):
        UVMSequenceItem.__init__(self, name)
        self.addr = 0
        self.rand("addr", range(0, 65536 - 1))
        self.read_write = 0
        self.rand("read_write", [READ, WRITE])
        self.size = 0
        self.rand("size", [1, 2, 4, 8])
        self.data = []
        #self.rand("data")
        self.wait_state = []
        #self.rand("wait_state")
        self.error_pos = 0
        #c_data_wait_size = lambda data, size: len(data) == size
        #self.constraint(c_data_wait_size)
        #self.rand("error_pos")
        self.transmit_delay = 0
        self.rand("transmit_delay", range(0, 11))
        self.master = ""
        self.slave = ""

    def post_randomize(self):
        self.wait_state = []
        self.data = []
        for i in range(self.size):
            self.wait_state.append(sv.urandom_range(0, 1))
            self.data.append(sv.urandom_range(1, 1 << 8))


    def do_copy(self, rhs):
        self.addr = rhs.addr
        self.read_write = rhs.read_write
        self.size = rhs.size
        for val in rhs.data:
            self.data.append(val)
        for val in rhs.wait_state:
            self.wait_state.append(val)

    def do_clone(self):
        new_obj = ubus_transfer()
        new_obj.copy(self)
        return new_obj

    def convert2string(self):
        rw_str = "READ"
        if self.read_write == WRITE:
            rw_str = "WRITE"
        elif self.read_write == NOP:
            rw_str = "NOP"
        res = "ID: " + str(self.m_transaction_id)
        res += ", addr: " + str(self.addr) + ", read_write: " + rw_str
        res += ", size: " + str(self.size) + ", wait_state: " + str(self.wait_state)
        res += "\ndata: "
        for d in self.data:
            res += hex(d) + ','
        res += "\n"
        return res


uvm_object_utils(ubus_transfer)

# TODO implement the field macros
#  `uvm_object_utils_begin(ubus_transfer)
#    `uvm_field_int      (addr,           UVM_DEFAULT)
#    `uvm_field_enum     (ubus_read_write_enum, read_write, UVM_DEFAULT)
#    `uvm_field_int      (size,           UVM_DEFAULT)
#    `uvm_field_array_int(data,           UVM_DEFAULT)
#    `uvm_field_array_int(wait_state,     UVM_DEFAULT)
#    `uvm_field_int      (error_pos,      UVM_DEFAULT)
#    `uvm_field_int      (transmit_delay, UVM_DEFAULT)
#    `uvm_field_string   (master,         UVM_DEFAULT|UVM_NOCOMPARE)
#    `uvm_field_string   (slave,          UVM_DEFAULT|UVM_NOCOMPARE)
#  `uvm_object_utils_end



#//
#//----------------------------------------------------------------------
#//   Copyright 2007-2010 Mentor Graphics Corporation
#//   Copyright 2007-2011 Cadence Design Systems, Inc.
#//   Copyright 2010-2011 Synopsys, Inc.
#//   Copyright 2019-2020 Tuomas Poikela (tpoikela)
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

import cocotb
from cocotb.triggers import *

from uvm.base.uvm_resource_db import UVMResourceDb
from uvm.base.sv import sv_if, sv

from uvm.base.uvm_component import *
from uvm.comps.uvm_env import *
from uvm.base.uvm_resource_db import *
from uvm import run_test



# About: uvm_exmples/mechanism/interfaces
# This example will illustrate how to create a pin interface and mod port interfaces for a simple dut.
# Connect a component "driver" to the pin interfaces, then to the dut.


#//----------------------------------------------------------------------
#// interface mem_pins_if
#//----------------------------------------------------------------------

class pin_if(sv_if):

    def __init__(self, dut):
        self.signals = ['clk', 'address', 'wr_data', 'rd_data', 'rst', 'rw',
            'req', 'ack', 'err']
        bus_map = {}
        for sig in self.signals:
            bus_map[sig] = sig
        super().__init__(dut, '', bus_map)
        self.slave_mp = self
        self.master_mp = self
        self.monitor_mp = self
        #  bit [15:0] address
        #  bit [7:0]  wr_data
        #  bit [7:0] rd_data
        #  bit rst
        #  bit rw
        #  bit req
        #  bit ack
        #  bit err

        #  modport master_mp(
        #   input  clk,
        #   input  rst,
        #   output address,
        #   output wr_data,
        #   input  rd_data,
        #   output req,
        #   output rw,
        #   input  ack,
        #   input  err );

        #  modport slave_mp(
        #   input  clk,
        #   input  rst,
        #   input  address,
        #   input  wr_data,
        #   output rd_data,
        #   input  req,
        #   input  rw,
        #   output ack,
        #   output err );

        #  modport monitor_mp(
        #   input  clk,
        #   input  rst,
        #   input  address,
        #   input  wr_data,
        #   input  rd_data,
        #   input  req,
        #   input  rw ,
        #   input  ack,
        #   input  err )


#//---------------------------------------------------------------------
#// component driver
#//----------------------------------------------------------------------


class driver(UVMComponent):

    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.pif = None

    def connect_phase(self, phase):
        arr = []
        sv.sv_assert(UVMResourceDb.read_by_name(self.get_full_name(), 'pif', arr))
        self.pif = arr[0]

    
    async def run_phase(self, phase):
        while True:
            await FallingEdge(self.pif.clk)
            uvm_info("driver", "posedge clk", UVM_NONE)


#//----------------------------------------------------------------------
#// environment env
#//----------------------------------------------------------------------

class env(UVMEnv):
    #
    #  local pin_vif pif
    #  driver d
    #
    def __init__(self, name, parent=None):
        super().__init__(name, parent)
        self.d = driver("driver", self)
        self.pif = None


    
    async def run_phase(self, phase):
        phase.raise_objection(self)
        await Timer(100, "NS")
        phase.drop_objection(self)



#//----------------------------------------------------------------------
#// module clkgen
#//----------------------------------------------------------------------


async def module_clkgen(clk):
    while True:
        await Timer(5, "NS")
        clk <= 1
        await Timer(5, "NS")
        clk <= 0


#//----------------------------------------------------------------------
#// module top
#//----------------------------------------------------------------------

@cocotb.test()
async def module_top(dut):

    pif = pin_if(dut) #  pin_if pif(clk)
    ck_proc = cocotb.fork(module_clkgen(pif.clk))
    #ck = clkgen(pif.clk) #  clkgen ck(clk)

    #d = dut(pif.slave_mp)

    e = env("e_env")
    UVMResourceDb.set("env.driver", "pif", pif)
    UVMResourceDb.dump()
    await run_test()



#
#------------------------------------------------------------------------------
#   Copyright 2007-2011 Mentor Graphics Corporation
#   Copyright 2007-2011 Cadence Design Systems, Inc.
#   Copyright 2010 Synopsys, Inc.
#   Copyright 2014 NVIDIA Corporation
#   All Rights Reserved Worldwide
#
#   Licensed under the Apache License, Version 2.0 (the
#   "License"); you may not use this file except in
#   compliance with the License.  You may obtain a copy of
#   the License at
#
#       http://www.apache.org/licenses/LICENSE-2.0
#
#   Unless required by applicable law or agreed to in
#   writing, software distributed under the License is
#   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#   CONDITIONS OF ANY KIND, either express or implied.  See
#   the License for the specific language governing
#   permissions and limitations under the License.
#------------------------------------------------------------------------------


"""
The :class:`UVMVoid` class is the base class for all UVM classes. It is an abstract
class with no data members or functions. It allows for generic containers of
objects to be created, similar to a void pointer in the C programming
language. User classes derived directly from :class:`UVMVoid` inherit none of the
UVM functionality, but such classes may be placed in :class:`UVMVoid`-typed
containers along with other UVM objects.
"""

from typing import Dict, Optional, Any

import re
from ..macros.uvm_message_defines import uvm_error, uvm_warning
from .uvm_object_globals import (UVM_BIN, UVM_COMPARE, UVM_COPY, UVM_DEC, UVM_FLAGS, UVM_NONE,
                                 UVM_NORADIX, UVM_OCT, UVM_PACK, UVM_PRINT, UVM_RECORD, UVM_SETINT,
                                 UVM_SETOBJ, UVM_SETSTR, UVM_STRING, UVM_TIME, UVM_UNPACK,
                                 UVM_UNSIGNED, UVM_HEX)
from .uvm_scope_stack import UVMScopeStack
from .sv import sv


class UVMVoid(object):
    pass


UVM_APPEND = 0
UVM_PREPEND = 1

UVM_ENABLE_FIELD_CHECKS = 0


def isunknown(value):
    return False


def uvm_get_array_index_string(arg: str, is_wildcard: int) -> str:
    """
    Function- uvm_get_array_index_string
    Args:
    Returns:
    """
    i = 0
    res = ""
    is_wildcard = 1
    i = len(arg) - 1
    if(arg[i] == "]"):
        while (i > 0 and (arg[i] != "[")):
            if ((arg[i] == "*") or (arg[i] == "?")):
                i = 0
            i -= 1

    if i > 0:
        res = arg[i+1: len(arg)-1]
        is_wildcard = 0
    return res


def uvm_bitstream_to_string(value, size, radix=UVM_NORADIX, radix_str="") -> str:
    """
    Function- uvm_bitstream_to_string

    Args:
        value (int):
        size:
        radix:
        radix_str:
    Returns:
        str: Given number as string.
    """
    # sign extend & don't show radix for negative values
    if radix == UVM_DEC and (value >> (size-1)):
        return "{}".format(value)

    # TODO $countbits(value,'z) would be even better
    if isunknown(value):
        _t = 0
        for idx in range(0, size):
            _t[idx] = value[idx]
            value = _t
    else:
        value &= (1 << size) - 1
    return num_with_radix(radix, radix_str, value)


def uvm_integral_to_string(value, size, radix=UVM_NORADIX, radix_str="") -> str:
    """
    Function- uvm_integral_to_string

    Args:
        value (int):
        size (int):
        radix:
        radix_str (str):
    Returns:
        str:
    """
    # sign extend & don't show radix for negative values
    if radix == UVM_DEC and (value >> (size-1)) == 1:
        return "{}".format(value)

    # TODO $countbits(value,'z) would be even better
    if isunknown(value) is True:
        _t = 0
        for idx in range(0, size):
            _t[idx] = value[idx]
            value = _t
    else:
        value &= (1 << size) - 1
    return num_with_radix(radix, radix_str, value)


def uvm_object_value_str(v) -> str:
    """
    Function- uvm_object_value_str

    Args:
        v (object): Object to convert to string.
    Returns:
        str: Inst ID for `UVMObject`, otherwise uses str()
    """
    if v is None:
        return "<null>"
    res = ""
    if hasattr(v, 'get_inst_id'):
        res = "{}".format(v.get_inst_id())
        res = "@" + res
    else:
        res = str(v)
    return res


def uvm_leaf_scope(full_name: str, scope_separator=".") -> str:
    """
    Function- uvm_leaf_scope

    Args:
        full_name (str):
        scope_separator (str):
    Returns:
        str: Leaf scope name without any brackets/indexing.
    """
    bmatches = 0

    bracket_match = get_bracket_match(scope_separator)

    # Only use bracket matching if the input string has the end match
    if bracket_match != "" and bracket_match != full_name[-1]:
        bracket_match = ""

    if bracket_match == "":
        regex = re.compile(scope_separator)
        if scope_separator == ".":
            regex = re.compile(r"\.")
        if not regex.search(full_name):
            return full_name

    pos = 0
    for pos in range(len(full_name) - 1, 0, -1):
        if full_name[pos] == bracket_match:
            bmatches += 1
        elif full_name[pos] == scope_separator:
            bmatches -= 1
            if bmatches == 0 or bracket_match == "":
                break

    res = ""
    if pos > 0:
        max_pos = len(full_name)
        if scope_separator != ".":
            pos -= 1
            max_pos -= 1
            if full_name[pos+1] == scope_separator:
                pos += 1
        res = full_name[pos+1:max_pos]
    else:
        res = full_name
    return res


def get_bracket_match(scope_separator: str) -> str:
    bracket_match = ""
    if scope_separator == "[":
        bracket_match = "]"
    elif scope_separator == "(":
        bracket_match = ")"
    elif scope_separator == "<":
        bracket_match = ">"
    elif scope_separator == "{":
        bracket_match = "}"
    return bracket_match


def num_with_radix(radix, radix_str: str, value) -> str:
    if radix == UVM_BIN:
        return "{}{:b}".format(radix_str, value)
    if radix == UVM_OCT:
        return "{}{}".format(radix_str, value)
    if radix == UVM_UNSIGNED:
        return "{}{}".format(radix_str, value)
    if radix == UVM_STRING:
        return "{}{}".format(radix_str, value)
    if radix == UVM_TIME:
        return "{}{}".format(radix_str, value)
    if radix == UVM_DEC:
        return "{}{}".format(radix_str, value)
    if radix == UVM_HEX:
        return "{}{:X}".format(radix_str, value)
    return "{}{}".format(radix_str, value)


class ProcessContainerC:
    def __init__(self, p_):
        self.p = p_


class UVMStatusContainer:
    """
    Internal class to contain status information for automation methods.
    """

    #  static bit field_array[string];
    field_array: Dict[str, bool] = {}
    #  static bit print_matches;
    print_matches = False

    def __init__(self):
        """
         The clone setting is used by the set/get config to know if cloning is on.
        """
        self.clone = True
        # Information variables used by the macro functions for storage.
        self.warning = False
        self.status = False
        self.bitstream = 0
        self.intv = 0
        self.element = 0
        self.stringv = ""
        self.scratch1 = ""
        self.scratch2 = ""
        self.key = ""
        self.object = None
        self.array_warning_done = False
        # The scope stack is used for messages that are emitted by policy classes.
        self.scope = UVMScopeStack()

        self.m_uvm_cycle_scopes = []  # uvm_object [$];

        #  //Used for checking cycles. When a data function is entered, if the depth is
        #  //non-zero, then then the existeance of the object in the map means that a
        #  //cycle has occured and the function should immediately exit. When the
        #  //function exits, it should reset the cycle map so that there is no memory
        #  //leak.
        self.cycle_check = {}  # bit cycle_check[uvm_object];

        #  //These are the policy objects currently in use. The policy object gets set
        #  //when a function starts up. The macros use this.
        # self.comparer = None  # UVMComparer
        # self.packer = None  # UVMPacker
        # self.recorder = None  # UVMRecorder
        # self.printer = None  # UVMPrinter

    def do_field_check(self, field, obj):
        if UVM_ENABLE_FIELD_CHECKS:
            if field in self.field_array:
                uvm_error("MLTFLD", "Field {} is defined multiple times in type '{}'".
                    format(field, obj.get_type_name()))
        UVMStatusContainer.field_array[field] = 1

    def get_function_type(self, what: int) -> str:
        if what == UVM_COPY:
            return "copy"
        elif what == UVM_COMPARE:
            return "compare"
        elif what == UVM_PRINT:
            return "print"
        elif what == UVM_RECORD:
            return "record"
        elif what == UVM_PACK:
            return "pack"
        elif what == UVM_UNPACK:
            return "unpack"
        elif what == UVM_FLAGS:
            return "get_flags"
        elif what == UVM_SETINT:
            return "set"
        elif what == UVM_SETOBJ:
            return "set_object"
        elif what == UVM_SETSTR:
            return "set_string"
        else:
            return "unknown"

    def get_full_scope_arg(self):
        """

        Returns:
        """
        return self.scope.get()



    #  // utility function used to perform a cycle check when config setting are pushed
    #  // to uvm_objects. the function has to look at the current object stack representing
    #  // the call stack of all __m_uvm_field_automation() invocations.
    #  // it is a only a cycle if the previous __m_uvm_field_automation call scope
    #  // is not identical with the current scope AND the scope is already present in the
    #  // object stack
    #  uvm_object m_uvm_cycle_scopes[$];

    def m_do_cycle_check(self, scope):
        ll = None
        #if m_uvm_cycle_scopes.size() == 0 ? null : m_uvm_cycle_scopes[$];
        if len(self.m_uvm_cycle_scopes) > 0:
            ll = self.m_uvm_cycle_scopes

        # we have been in this scope before (but actually right before so assuming a super/derived
        # context of the same object)
        if ll == scope:
            self.m_uvm_cycle_scopes.append(scope)
            return 0
        else:
            # now check if we have already been in this scope before
            idx = -1
            #uvm_object m[$] = m_uvm_cycle_scopes.find_first(item) with (item == scope);
            for i in range(len(self.m_uvm_cycle_scopes)):
                if self.m_uvm_cycle_scopes[i] == scope:
                    idx = i
                    break
            if idx != -1:
                return 1  # detected a cycle
            else:
                self.m_uvm_cycle_scopes.append(scope)
                return 0


def m_uvm_string_queue_join(i):
    res = ""
    for idx in range(len(i)):
        res = res + i[idx]
    return res


#//------------------------------------------------------------------------------
#// CLASS: uvm_utils #(TYPE,FIELD)
#//
#// This class contains useful template functions.
#//
#//------------------------------------------------------------------------------

class UVMUtils():  # (type TYPE=int, string FIELD="config")


    @classmethod
    def find_all(cls, start, TYPE):
        """
        Function: find_all

        Recursively finds all component instances of the parameter type `TYPE`,
        starting with the component given by `start`. Uses <uvm_root::find_all>.

        Args:
            start:
            TYPE:
        Returns:
        """
        from .uvm_coreservice import UVMCoreService
        comp_list = []
        types = []
        cs = UVMCoreService.get()
        top = cs.get_root()
        top.find_all("*", comp_list, start)
        for comp in comp_list:
            typ = []
            if sv.cast(typ, comp, TYPE):
                types.append(typ[0])

        if len(types) == 0:
            uvm_warning("find_type-no match", "Instance of type '" + TYPE.type_name
                + " not found in component hierarchy beginning at " + start.get_full_name())
        return types


    #  static def find(self,uvm_component start):
    #    types_t types = find_all(start)
    #    if (types.size() == 0)
    #      return None
    #    if (types.size() > 1):
    #      uvm_warning("find_type-multi match",{"More than one instance of type '",TYPE::type_name,
    #         " found in component hierarchy beginning at ",start.get_full_name()})
    #      return None
    #    end
    #    return types[0]
    #  endfunction
    #

    @classmethod
    def create_type_by_name(cls, type_name, contxt):
        from .uvm_coreservice import UVMCoreService
        obj = None
        typ = None  # TYPE
        cs = UVMCoreService.get()
        factory = cs.get_factory()

        obj = factory.create_object_by_name(type_name,contxt,type_name)
        typ = obj
        # TODO cast
        #    if (!sv.cast(typ,obj))
        #     uvm_error("WRONG_TYPE",{"The type_name given '",type_name,
        #            "' with context '",contxt,"' did not produce the expected type."})
        return typ


    @classmethod
    def get_config(cls, comp, is_fatal):
        """
        This method gets the object config of type `TYPE`
        associated with component `comp`.
        We check for the two kinds of error which may occur with this kind of
        operation.

        Args:
            cls:
            comp:
            is_fatal:
        Returns:
        """
        obj = None
        cfg = None
        # TODO
        #
        #    if (!m_uvm_config_obj_misc::get(comp,"",FIELD, obj)):
        #      if (is_fatal)
        #        comp.uvm_report_fatal("NO_SET_CFG", {"no set_config to field '", FIELD,
        #                           "' for component '",comp.get_full_name(),"'"},
        #                           UVM_MEDIUM, `uvm_file , `uvm_line  )
        #      else
        #        comp.uvm_report_warning("NO_SET_CFG", {"no set_config to field '", FIELD,
        #                           "' for component '",comp.get_full_name(),"'"},
        #                           UVM_MEDIUM, `uvm_file , `uvm_line  )
        #      return None
        #    end
        #
        #    if (!sv.cast(cfg, obj)):
        #      if (is_fatal)
        #        comp.uvm_report_fatal( "GET_CFG_TYPE_FAIL",
        #                          {"set_config_object with field name ",FIELD,
        #                          " is not of type '",TYPE::type_name,"'"},
        #                          UVM_NONE , `uvm_file , `uvm_line )
        #      else
        #        comp.uvm_report_warning( "GET_CFG_TYPE_FAIL",
        #                          {"set_config_object with field name ",FIELD,
        #                          " is not of type '",TYPE::type_name,"'"},
        #                          UVM_NONE , `uvm_file , `uvm_line )
        #    end
        #
        return cfg



#//----------------------------------------------------------------------
#//   Copyright 2010-2011 Mentor Graphics Corporation
#//   Copyright 2011 Synopsys, Inc.
#//   Copyright 2013 Cadence Design Inc
#//   Copyright 2014 NVIDIA Corporation
#//   All Rights Reserved Worldwide
#//
#//   Licensed under the Apache License, Version 2.0 (the
#//   "License"); you may not use this file except in
#//   compliance with the License.  You may obtain a copy of
#//   the License at
#//
#//       http://www.apache.org/licenses/LICENSE-2.0
#//
#//   Unless required by applicable law or agreed to in
#//   writing, software distributed under the License is
#//   distributed on an "AS IS" BASIS, WITHOUT WARRANTIES OR
#//   CONDITIONS OF ANY KIND, either express or implied.  See
#//   the License for the specific language governing
#//   permissions and limitations under the License.
#//----------------------------------------------------------------------

#typedef class uvm_sequence_library_cfg;

#//
#// CLASS: uvm_sequence_library
#//
#// The ~uvm_sequence_library~ is a sequence that contains a list of registered
#// sequence types. It can be configured to create and execute these sequences
#// any number of times using one of several modes of operation, including a
#// user-defined mode.
#//
#// When started (as any other sequence), the sequence library will randomly
#// select and execute a sequence from its ~sequences~ queue. If in
#// <UVM_SEQ_LIB_RAND> mode, its <select_rand> property is randomized and used
#// as an index into ~sequences~.  When in <UVM_SEQ_LIB_RANDC> mode, the
#// <select_randc> property is used. When in <UVM_SEQ_LIB_ITEM> mode, only
#// sequence items of the ~REQ~ type are generated and executed--no sequences
#// are executed. Finally, when in <UVM_SEQ_LIB_USER> mode, the
#// <select_sequence> method is called to obtain the index for selecting the
#// next sequence to start. Users can override this method in subtypes to
#// implement custom selection algorithms.
#//
#// Creating a subtype of a sequence library requires invocation of the
#// <`uvm_sequence_library_utils> macro in its declaration and calling
#// the <init_sequence_library> method in its constructor. The macro
#// and function are needed to populate the sequence library with any
#// sequences that were statically registered with it or any of its base
#// classes.
#//
#//| class my_seq_lib extends uvm_sequence_library #(my_item);
#//|   `uvm_object_utils(my_seq_lib)
#//|   `uvm_sequence_library_utils(my_seq_lib)
#//|    function new(string name="");
#//|      super.new(name);
#//|      init_sequence_library();
#//|    endfunction
#//|    ...
#//| endclass
#//
#//------------------------------------------------------------------------------

#class uvm_sequence_library #(type REQ=uvm_sequence_item,RSP=REQ) extends uvm_sequence #(REQ,RSP);
#
#   // Function: new
#   //
#   // Create a new instance of this class
#   //
#   extern function new(string name="");
#
#
#   // Function: get_type_name
#   //
#   // Get the type name of this class
#   //
#   extern virtual function string get_type_name();
#
#
#
#   //--------------------------
#   // Group: Sequence selection
#   //--------------------------
#
#   // Variable: selection_mode
#   //
#   // Specifies the mode used to select sequences for execution
#   //
#   // If you do not have access to an instance of the library,
#   // use the configuration resource interface.
#   //
#   // The following example sets the ~config_seq_lib~ as the default
#   // sequence for the 'main' phase on the sequencer to be
#   // located at "env.agent.sequencer"
#   // and set the selection mode to <UVM_SEQ_LIB_RANDC>. If the
#   // settings are being done from within a component, the first
#   // argument must be ~this~ and the second argument a path
#   // relative to that component.
#   // 
#   //
#   //| uvm_config_db #(uvm_object_wrapper)::set(null,
#   //|                                    "env.agent.sequencer.main_phase",
#   //|                                    "default_sequence",
#   //|                                    main_seq_lib::get_type());
#   //|
#   //| uvm_config_db #(uvm_sequence_lib_mode)::set(null,
#   //|                                    "env.agent.sequencer.main_phase",
#   //|                                    "default_sequence.selection_mode",
#   //|                                    UVM_SEQ_LIB_RANDC);
#   //
#   // Alternatively, you may create an instance of the sequence library
#   // a priori, initialize all its parameters, randomize it, then set it
#   // to run as-is on the sequencer. 
#   //
#   //| main_seq_lib my_seq_lib;
#   //| my_seq_lib = new("my_seq_lib");
#   //|
#   //| my_seq_lib.selection_mode = UVM_SEQ_LIB_RANDC;
#   //| my_seq_lib.min_random_count = 500;
#   //| my_seq_lib.max_random_count = 1000;
#   //| void'(my_seq_lib.randomize());
#   //|
#   //| uvm_config_db #(uvm_sequence_base)::set(null,
#   //|                                    "env.agent.sequencer.main_phase",
#   //|                                    "default_sequence",
#   //|                                    my_seq_lib);
#   //|
#   //
#   uvm_sequence_lib_mode selection_mode;
#
#
#   // Variable: min_random_count
#   //
#   // Sets the minimum number of items to execute. Use the configuration
#   // mechanism to set. See <selection_mode> for an example.
#   //
#   int unsigned min_random_count=10;
#
#
#   // Variable: max_random_count
#   //
#   // Sets the maximum number of items to execute. Use the configuration
#   // mechanism to set. See <selection_mode> for an example.
#   //
#   //
#   int unsigned max_random_count=10;
#
#
#
#   // Variable: sequences_executed
#   //
#   // Indicates the number of sequences executed, not including the
#   // currently executing sequence, if any.
#   //
#   protected int unsigned sequences_executed;
#
#
#   // Variable: sequence_count
#   //
#   // Specifies the number of sequences to execute when this sequence
#   // library is started. If in <UVM_SEQ_LIB_ITEM> mode, specifies the
#   // number of sequence items that will be generated.
#   //
#   rand  int unsigned sequence_count = 10;
#
#
#   // Variable: select_rand
#   //
#   // The index variable that is randomized to select the next sequence
#   // to execute when in UVM_SEQ_LIB_RAND mode
#   //
#   // Extensions may place additional constraints on this variable.
#   //
#   rand  int unsigned select_rand;
#
#
#   // Variable: select_randc
#   //
#   // The index variable that is randomized to select the next sequence
#   // to execute when in UVM_SEQ_LIB_RANDC mode
#   //
#   // Extensions may place additional constraints on this variable.
#   //
#   randc bit [15:0] select_randc;
#
#
#
#   // Variable- seqs_distrib
#   //
#   //
#   //
#   protected int seqs_distrib[string]  = '{default:0};
#
#
#   // Variable- sequences
#   //
#   // The container of all registered sequence types. For <sequence_count>
#   // times, this sequence library will randomly select and execute a
#   // sequence from this list of sequence types.
#   //
#   protected uvm_object_wrapper sequences[$];
#
#
#
#   // Constraint: valid_rand_selection
#   //
#   // Constrains <select_rand> to be a valid index into the ~sequences~ array
#   //
#   constraint valid_rand_selection {
#         select_rand inside {[0:sequences.size()-1]};
#   }
#
#
#
#   // Constraint: valid_randc_selection
#   //
#   // Constrains <select_randc> to be a valid index into the ~sequences~ array
#   //
#   constraint valid_randc_selection {
#         select_randc inside {[0:sequences.size()-1]};
#   }
#
#
#   // Constraint: valid_sequence_count
#   //
#   // Constrains <sequence_count> to lie within the range defined by
#   // <min_random_count> and <max_random_count>.
#   //
#   constraint valid_sequence_count {
#      sequence_count inside {[min_random_count:max_random_count]};
#   }
#
#
#
#   // Function: select_sequence
#   //
#   // Generates an index used to select the next sequence to execute. 
#   // Overrides must return a value between 0 and ~max~, inclusive.
#   // Used only for <UVM_SEQ_LIB_USER> selection mode. The
#   // default implementation returns 0, incrementing on successive calls,
#   // wrapping back to 0 when reaching ~max~.
#   //
#   extern virtual function int unsigned select_sequence(int unsigned max);
#
#
#
#   //-----------------------------
#   // Group: Sequence registration
#   //-----------------------------
#
#   // Function: add_typewide_sequence
#   //
#   // Registers the provided sequence type with this sequence library
#   // type. The sequence type will be available for selection by all instances
#   // of this class. Sequence types already registered are silently ignored.
#   //
#   extern static function void add_typewide_sequence(uvm_object_wrapper seq_type);
#
#
#
#   // Function: add_typewide_sequences
#   //
#   // Registers the provided sequence types with this sequence library
#   // type. The sequence types will be available for selection by all instances
#   // of this class. Sequence types already registered are silently ignored.
#   //
#   extern static function void add_typewide_sequences(uvm_object_wrapper seq_types[$]);
#
#
#   // Function: add_sequence
#   //
#   // Registers the provided sequence type with this sequence library
#   // instance. Sequence types already registered are silently ignored.
#   //
#   extern function void add_sequence(uvm_object_wrapper seq_type);
#
#
#   // Function: add_sequences
#   //
#   // Registers the provided sequence types with this sequence library
#   // instance. Sequence types already registered are silently ignored.
#   //
#   extern virtual function void add_sequences(uvm_object_wrapper seq_types[$]);
#
#
#   // Function: remove_sequence
#   //
#   // Removes the given sequence type from this sequence library
#   // instance. If the type was registered statically, the sequence queues of
#   // all instances of this library will be updated accordingly.
#   // A warning is issued if the sequence is not registered.
#   //
#   extern virtual function void remove_sequence(uvm_object_wrapper seq_type);
#
#
#   // Function: get_sequences
#   //
#   // 
#   // Append to the provided ~seq_types~ array the list of registered ~sequences~.
#   //
#   extern virtual function void get_sequences(ref uvm_object_wrapper seq_types[$]);
#
#
#   // Function: init_sequence_library
#   //
#   // All subtypes of this class must call init_sequence_library in its
#   // constructor.
#   extern function void init_sequence_library();
#
#   // Macro: uvm_sequence_library_utils
#   //
#   // All subtypes of this class must invoke the `uvm_sequence_library_utils
#   // macro.
#   //
#   //| class my_seq_lib extends uvm_sequence_library #(my_item);
#   //|   `uvm_object_utils(my_seq_lib)
#   //|   `uvm_sequence_library_utils(my_seq_lib)
#   //|    function new(string name="");
#   //|      super.new(name);
#   //|      init_sequence_library();
#   //|    endfunction
#   //|    ...
#   //| endclass
#
#   //------------------------------------------
#   // PRIVATE - INTERNAL - NOT PART OF STANDARD
#   //------------------------------------------
#
#   `uvm_object_param_utils(uvm_sequence_library #(REQ,RSP))
#   typedef uvm_sequence_library #(REQ,RSP) this_type;
#
#   static const string type_name = "uvm_sequence_library #(REQ,RSP)";
#   static protected uvm_object_wrapper m_typewide_sequences[$];
#   bit m_abort;
#
#   extern static   function bit  m_static_check(uvm_object_wrapper seq_type);
#   extern static   function bit  m_check(uvm_object_wrapper seq_type, this_type lib);
#   extern          function bit  m_dyn_check(uvm_object_wrapper seq_type);
#   extern          function void m_get_config();
#   extern static   function bit  m_add_typewide_sequence(uvm_object_wrapper seq_type);
#   extern virtual  task          execute(uvm_object_wrapper wrap);
#
#   extern virtual  task          body();
#   extern virtual  function void do_print(uvm_printer printer);
#   extern          function void pre_randomize();
#
#endclass
#
#
#
#//------------------------------------------------------------------------------
#//
#// Class: uvm_sequence_library_cfg
#//
#// A convenient container class for configuring all the sequence library
#// parameters using a single ~set~ command.
#//
#//| uvm_sequence_library_cfg cfg;
#//| cfg = new("seqlib_cfg", UVM_SEQ_LIB_RANDC, 1000, 2000);
#//|
#//| uvm_config_db #(uvm_sequence_library_cfg)::set(null,
#//|                                    "env.agent.sequencer.main_ph",
#//|                                    "default_sequence.config",
#//|                                    cfg);
#//|
#//------------------------------------------------------------------------------
#
#class uvm_sequence_library_cfg extends uvm_object;
#  `uvm_object_utils(uvm_sequence_library_cfg)
#  uvm_sequence_lib_mode selection_mode;
#  int unsigned min_random_count;
#  int unsigned max_random_count;
#  function new(string name="",
#               uvm_sequence_lib_mode mode=UVM_SEQ_LIB_RAND,
#               int unsigned min=1,
#               int unsigned max=10);
#    super.new(name);
#    selection_mode = mode;
#    min_random_count = min;
#    max_random_count = max;
#  endfunction
#endclass
#
#
#
#//------------------------------------------------------------------------------
#// IMPLEMENTATION
#//------------------------------------------------------------------------------
#
#// new
#// ---
#
#function uvm_sequence_library::new(string name="");
#   super.new(name);
#   init_sequence_library();
#   valid_rand_selection.constraint_mode(0);
#   valid_randc_selection.constraint_mode(0);
#endfunction
#
#
#// get_type_name
#// -------------
#
#function string uvm_sequence_library::get_type_name();
#  return type_name;
#endfunction
#
#
#// m_add_typewide_sequence
#// -----------------------
#
#function bit uvm_sequence_library::m_add_typewide_sequence(uvm_object_wrapper seq_type);
#  this_type::add_typewide_sequence(seq_type);
#  return 1;
#endfunction
#
#
#// add_typewide_sequence
#// ---------------------
#
#function void uvm_sequence_library::add_typewide_sequence(uvm_object_wrapper seq_type);
#  if (m_static_check(seq_type))
#    m_typewide_sequences.push_back(seq_type);
#endfunction
#
#
#// add_typewide_sequences
#// ----------------------
#
#function void uvm_sequence_library::add_typewide_sequences(uvm_object_wrapper seq_types[$]);
#  foreach (seq_types[i])
#    add_typewide_sequence(seq_types[i]);
#endfunction
#
#
#// add_sequence
#// ------------
#
#function void uvm_sequence_library::add_sequence(uvm_object_wrapper seq_type);
#  if (m_dyn_check(seq_type))
#    sequences.push_back(seq_type);
#endfunction
#
#
#// add_sequences
#// -------------
#
#function void uvm_sequence_library::add_sequences(uvm_object_wrapper seq_types[$]);
#  foreach (seq_types[i])
#    add_sequence(seq_types[i]);
#endfunction
#
#
#// remove_sequence
#// ---------------
#
#function void uvm_sequence_library::remove_sequence(uvm_object_wrapper seq_type);
#  foreach (sequences[i])
#    if (sequences[i] == seq_type) begin
#      sequences.delete(i);
#      return;
#    end
#endfunction
#
#
#// get_sequences
#// -------------
#
#function void uvm_sequence_library::get_sequences(ref uvm_object_wrapper seq_types[$]);
#  foreach (sequences[i])
#    seq_types.push_back(sequences[i]);
#endfunction
#
#
#// select_sequence
#// ---------------
#
#function int unsigned uvm_sequence_library::select_sequence(int unsigned max);
#  static int unsigned counter;
#  select_sequence = counter;
#  counter++;
#  if (counter >= max)
#    counter = 0;
#endfunction
#
#
#//----------//
#// INTERNAL //
#//----------//
#
#
#// init_sequence_library
#// ---------------------
#
#function void uvm_sequence_library::init_sequence_library();
#  foreach (this_type::m_typewide_sequences[i])
#    sequences.push_back(this_type::m_typewide_sequences[i]);
#endfunction
#
#
#
#// m_static_check
#// --------------
#
#
#function bit uvm_sequence_library::m_static_check(uvm_object_wrapper seq_type);
#  if (!m_check(seq_type,null))
#    return 0;
#  foreach (m_typewide_sequences[i])
#    if (m_typewide_sequences[i] == seq_type)
#      return 0;
#  return 1;
#endfunction
#
#
#// m_dyn_check
#// -----------
#
#function bit uvm_sequence_library::m_dyn_check(uvm_object_wrapper seq_type);
#  if (!m_check(seq_type,this))
#    return 0;
#  foreach (sequences[i])
#    if (sequences[i] == seq_type)
#      return 0;
#  return 1;
#endfunction
#
#
#// m_check
#// -------
#
#function bit uvm_sequence_library::m_check(uvm_object_wrapper seq_type, this_type lib);
#  uvm_object obj;
#  uvm_sequence_base seq;
#  uvm_root top;
#  uvm_coreservice_t cs;   
#  string name;
#  string typ;
#  obj = seq_type.create_object();
#  name = (lib == null) ? type_name : lib.get_full_name();
#  typ = (lib == null) ? type_name : lib.get_type_name();
#  cs = uvm_coreservice_t::get();   
#  top = cs.get_root();
#
#  if (!$cast(seq, obj)) begin
#    `uvm_error_context("SEQLIB/BAD_SEQ_TYPE",
#        {"Object '",obj.get_type_name(),
#        "' is not a sequence. Cannot add to sequence library '",name,
#        "'"},top)
#     return 0;
#  end
#  return 1;
#endfunction
#
#
#// pre_randomize
#// -------------
#
#function void uvm_sequence_library::pre_randomize();
#  m_get_config();
#endfunction
#
#
#// m_get_config
#// ------------
#
#function void uvm_sequence_library::m_get_config();
#
#  uvm_sequence_library_cfg cfg;
#  string phase_name;
#  uvm_phase starting_phase = get_starting_phase();
#   
#  if (starting_phase != null) begin
#    phase_name = {starting_phase.get_name(),"_phase"};
#  end
#  if (uvm_config_db #(uvm_sequence_library_cfg)::get(m_sequencer, 
#                                        phase_name,
#                                        "default_sequence.config",
#                                        cfg) ) begin
#    selection_mode = cfg.selection_mode; 
#    min_random_count = cfg.min_random_count; 
#    max_random_count = cfg.max_random_count; 
#  end
#  else begin
#    void'(uvm_config_db #(int unsigned)::get(m_sequencer, 
#                                        phase_name,
#                                        "default_sequence.min_random_count",
#                                        min_random_count) );
#
#    void'(uvm_config_db #(int unsigned)::get(m_sequencer, 
#                                        phase_name,
#                                        "default_sequence.max_random_count",
#                                        max_random_count) );
#
#    void'(uvm_config_db #(uvm_sequence_lib_mode)::get(m_sequencer, 
#                                        phase_name,
#                                        "default_sequence.selection_mode",
#                                        selection_mode) );
#  end
#
#  if (max_random_count == 0) begin
#    `uvm_warning("SEQLIB/MAX_ZERO",
#       $sformatf("max_random_count (%0d) zero. Nothing will be done.",
#       max_random_count))
#    if (min_random_count > max_random_count)
#      min_random_count = max_random_count;
#  end
#  else if (min_random_count > max_random_count) begin
#    `uvm_error("SEQLIB/MIN_GT_MAX",
#       $sformatf("min_random_count (%0d) greater than max_random_count (%0d). Setting min to max.",
#       min_random_count,max_random_count))
#    min_random_count = max_random_count;
#  end
#  else begin
#    if (selection_mode == UVM_SEQ_LIB_ITEM) begin
#      uvm_sequencer #(REQ,RSP) seqr;
#      uvm_object_wrapper lhs = REQ::get_type();
#      uvm_object_wrapper rhs = uvm_sequence_item::get_type();
#      if (lhs == rhs) begin
#        `uvm_error("SEQLIB/BASE_ITEM", {"selection_mode cannot be UVM_SEQ_LIB_ITEM when ",
#          "the REQ type is the base uvm_sequence_item. Using UVM_SEQ_LIB_RAND mode"})
#        selection_mode = UVM_SEQ_LIB_RAND;
#      end
#      if (m_sequencer == null || !$cast(seqr,m_sequencer)) begin
#        `uvm_error("SEQLIB/VIRT_SEQ", {"selection_mode cannot be UVM_SEQ_LIB_ITEM when ",
#          "running as a virtual sequence. Using UVM_SEQ_LIB_RAND mode"})
#        selection_mode = UVM_SEQ_LIB_RAND;
#      end
#    end
#  end
#
#endfunction
#
#
#// body
#// ----
#
#task uvm_sequence_library::body();
#
#  uvm_object_wrapper wrap;
#  uvm_phase starting_phase = get_starting_phase();
#   
#  if (m_sequencer == null) begin
#    `uvm_fatal("SEQLIB/VIRT_SEQ", {"Sequence library 'm_sequencer' handle is null; ",
#      " no current support for running as a virtual sequence."})
#     return;
#  end
#
#  if (sequences.size() == 0) begin
#    `uvm_error("SEQLIB/NOSEQS", "Sequence library does not contain any sequences. Did you forget to call init_sequence_library() in the constructor?")
#    return;
#  end
#
#  if (do_not_randomize)
#    m_get_config();
#
#  m_safe_raise_starting_phase({"starting sequence library ",get_full_name()," (", get_type_name(),")"});
#
#  `uvm_info("SEQLIB/START",
#     $sformatf("Starting sequence library %s in %s phase: %0d iterations in mode %s",
#      get_type_name(),
#      (starting_phase != null ? starting_phase.get_name() : "unknown"),
#      sequence_count, selection_mode.name()),UVM_LOW)
#
#   `uvm_info("SEQLIB/SPRINT",{"\n",sprint(uvm_default_table_printer)},UVM_FULL)
#
#    case (selection_mode)
#
#      UVM_SEQ_LIB_RAND: begin
#        valid_rand_selection.constraint_mode(1);
#        valid_sequence_count.constraint_mode(0);
#        for (int i=1; i<=sequence_count; i++) begin
#          if (!randomize(select_rand)) begin
#            `uvm_error("SEQLIB/RAND_FAIL", "Random sequence selection failed")
#            break;
#          end
#          else begin
#            wrap = sequences[select_rand];
#          end
#          execute(wrap);
#        end
#        valid_rand_selection.constraint_mode(0);
#        valid_sequence_count.constraint_mode(1);
#      end
#
#      UVM_SEQ_LIB_RANDC: begin
#        uvm_object_wrapper q[$];
#        valid_randc_selection.constraint_mode(1);
#        valid_sequence_count.constraint_mode(0);
#        for (int i=1; i<=sequence_count; i++) begin
#          if (!randomize(select_randc)) begin
#            `uvm_error("SEQLIB/RANDC_FAIL", "Random sequence selection failed")
#            break;
#          end
#          else begin
#            wrap = sequences[select_randc];
#          end
#          q.push_back(wrap);
#        end
#        valid_randc_selection.constraint_mode(0);
#        valid_sequence_count.constraint_mode(1);
#        foreach(q[i])
#          execute(q[i]);
#        valid_randc_selection.constraint_mode(0);
#        valid_sequence_count.constraint_mode(1);
#      end
#
#      UVM_SEQ_LIB_ITEM: begin
#        for (int i=1; i<=sequence_count; i++) begin
#          wrap = REQ::get_type();
#          execute(wrap);
#        end
#      end
#
#      UVM_SEQ_LIB_USER: begin
#        for (int i=1; i<=sequence_count; i++) begin
#          int user_selection;
#          user_selection = select_sequence(sequences.size()-1);
#          if (user_selection >= sequences.size()) begin
#            `uvm_error("SEQLIB/USER_FAIL", "User sequence selection out of range")
#            wrap = REQ::get_type();
#          end
#          else begin
#            wrap = sequences[user_selection];
#          end
#          execute(wrap);
#        end
#      end
#
#      default: begin
#        `uvm_fatal("SEQLIB/RAND_MODE", 
#           $sformatf("Unknown random sequence selection mode: %0d",selection_mode))
#      end
#     endcase
#
#  `uvm_info("SEQLIB/END",{"Ending sequence library in phase ",
#            (starting_phase != null ? starting_phase.get_name() : "unknown")},UVM_LOW)
# 
#  `uvm_info("SEQLIB/DSTRB",$sformatf("%p",seqs_distrib),UVM_HIGH)
#
#  m_safe_drop_starting_phase({"starting sequence library ",get_full_name()," (", get_type_name(),")"});
#
#endtask
#
#
#// execute
#// -------
#
#task uvm_sequence_library::execute(uvm_object_wrapper wrap);
#
#  uvm_object obj;
#  uvm_sequence_item seq_or_item;
#  uvm_sequence_base seq_base;
#  REQ req_item;
#  
#  uvm_coreservice_t cs = uvm_coreservice_t::get();                                                     
#  uvm_factory factory=cs.get_factory();
#
#  obj = factory.create_object_by_type(wrap,get_full_name(),
#           $sformatf("%s:%0d",wrap.get_type_name(),sequences_executed+1));
#
#  if (!$cast(seq_base, obj)) begin
#     // If we're executing an item (not a sequence)
#     if (!$cast(req_item, obj)) begin
#        // But it's not our item type (This can happen if we were parameterized with
#        // a pure virtual type, because we're getting get_type() from the base class)
#        `uvm_error("SEQLIB/WRONG_ITEM_TYPE", {"The item created by '", get_full_name(), "' when in 'UVM_SEQ_LIB_ITEM' mode doesn't match the REQ type which  was passed in to the uvm_sequence_library#(REQ[,RSP]), this can happen if the REQ type which was passed in was a pure-virtual type.  Either configure the factory overrides to properly generate items for this sequence library, or do not execute this sequence library in UVM_SEQ_LIB_ITEM mode."})
#         return;
#     end
#  end
#   
#  void'($cast(seq_or_item,obj)); // already qualified, 
#   
#  `uvm_info("SEQLIB/EXEC",{"Executing ",(seq_or_item.is_item() ? "item " : "sequence "),seq_or_item.get_name(),
#                           " (",seq_or_item.get_type_name(),")"},UVM_FULL)
#  seq_or_item.print_sequence_info = 1;
#  `uvm_rand_send(seq_or_item)
#  seqs_distrib[seq_or_item.get_type_name()] = seqs_distrib[seq_or_item.get_type_name()]+1;
#
#  sequences_executed++;
#
#endtask
#  
#
#
#// do_print
#// --------
#
#function void uvm_sequence_library::do_print(uvm_printer printer);
#   printer.print_field_int("min_random_count",min_random_count,32,UVM_DEC,,"int unsigned");
#   printer.print_field_int("max_random_count",max_random_count,32,UVM_DEC,,"int unsigned");
#   printer.print_generic("selection_mode","uvm_sequence_lib_mode",32,selection_mode.name());
#   printer.print_field_int("sequence_count",sequence_count,32,UVM_DEC,,"int unsigned");
#
#   printer.print_array_header("typewide_sequences",m_typewide_sequences.size(),"queue_object_types");
#   foreach (m_typewide_sequences[i])
#     printer.print_generic($sformatf("[%0d]",i),"uvm_object_wrapper","-",m_typewide_sequences[i].get_type_name());
#   printer.print_array_footer();
#
#   printer.print_array_header("sequences",sequences.size(),"queue_object_types");
#   foreach (sequences[i])
#     printer.print_generic($sformatf("[%0d]",i),"uvm_object_wrapper","-",sequences[i].get_type_name());
#   printer.print_array_footer();
#
#   printer.print_array_header("seqs_distrib",seqs_distrib.num(),"as_int_string");
#   foreach (seqs_distrib[typ]) begin
#     printer.print_field_int({"[",typ,"]"},seqs_distrib[typ],32,,UVM_DEC,"int unsigned");
#   end
#   printer.print_array_footer();
#endfunction
#
#
